"""Defines the Booking object"""

import requests

from fabman.fabman_object import FabmanObject


class Booking(FabmanObject):
    """
    Booking object as described in the Fabman API:
    https://fabman.io/api/v1/documentation#/bookings
    """

    def __str__(self):
        return f"Booking #{self.id}: {self.fromDateTime} - {self.untilDateTime}"

    def delete(self, **kwargs) -> requests.Response:
        """
        Deletes the booking. *WARNING: THIS CANNOT BE UNDONE.*

        :calls: "DELETE /bookings/{id}" \
		<https://fabman.io/api/v1/documentation#/bookings/deleteBookingsId>
  
        :returns: Response from the API with status code 204 (No Content) if successful
        :rtype: requests.Response
        """

        uri = f"/bookings/{self.id}"

        response = self._requester.request("DELETE", uri, _kwargs=kwargs)

        return response

    def update(self, **kwargs) -> None:
        """
        Update the booking object on the server. Returns the updated booking object.
        Updates all information in place.

        :calls: "PUT /bookings/{id}" \
		<https://fabman.io/api/v1/documentation#/bookings/putBookingsId>

        :returns: None
        :rtype: None
        """

        uri = f"/bookings/{self.id}"

        kwargs.update({"lockVersion": self.lockVersion})

        response = self._requester.request("PUT", uri, _kwargs=kwargs)

        data = response.json()

        for attr, val in data.items():
            setattr(self, attr, val)



"""Defines the Space class and other Space-related classes"""

from typing import Union

import requests

from fabman.embedded_list import EmbeddedList
from fabman.fabman_object import FabmanObject
from fabman.paginated_list import PaginatedList


class SpaceBillingSettings(FabmanObject):
    """Class for interacting with the space-billing-settings endpoint on the Fabman API"""

    def __str__(self):
        return f"SpaceBillingSettings for space #{self.space}"

    def delete_stripe(self, **kwargs) -> requests.Response:
        """
        Deletes Stripe information from a space. *WARNING: This is irreversible.*

        :calls: "DELETE /space/{id]/billing-settings/stripe" \
		<https://fabman.io/api/v1/documentation#/spaces/deleteSpacesIdBillingsettingsStripe>
  
        :return: response information of call
        :rtype: requests.Response
        """

        uri = f"/spaces/{self.space_id}/billing-settings/stripe"

        response = self._requester.request("DELETE", uri, _kwargs=kwargs)

        return response

    def update(self, **kwargs) -> None:
        """
        Updates the space-billing-settings. Attributes are updated in place with new information
        retrieved from the API.

        :calls: "PUT /space-billing-settings/{spaceBillingSettingsId}" \
		<https://fabman.io/api/v1/documentation#/space-billing-settings/putSpaceBillingSettingsSpaceBillingSettingsId>
  
        :return: None -- attributes are updated in place
        :rtype: None
        """

        uri = f"/spaces/{self.space_id}/billing-settings"

        kwargs.update({"lockVersion": self.lockVersion})

        response = self._requester.request("PUT", uri, _kwargs=kwargs)

        data = response.json()

        for attr, val in data.items():
            setattr(self, attr, val)

    def update_stripe(self, **kwargs) -> None:
        """
        Updates the space-billing-settings. Attributes are updated in place with new information
        retrieved from the API. Placeholder for future implementation

        :calls: "PUT /space-billing-settings/{spaceBillingSettingsId}" \
		<https://fabman.io/api/v1/documentation>
  
        :raises: NotImplementedError
        """

        raise NotImplementedError("This method is not yet implemented.")


class SpaceHoliday(FabmanObject):
    """Class for interacting with the space/{id}/holidays endpoints on the Fabman API"""

    def __str__(self):
        return f"SpaceHoliday #{self.id}: {self.title}"

    def delete(self, **kwargs) -> requests.Response:
        """
        Deletes a space holiday. *WARNING: This is irreversible.*

        :calls: "DELETE /space/{spaceId}/holidays/{holidayId}" \
		<https://fabman.io/api/v1/documentation#/spaces/deleteSpacesIdHolidaysHolidayid>
  
        :return: response information of call
        :rtype: requests.Response
        """

        uri = f"/spaces/{self.space_id}/holidays/{self.id}"

        response = self._requester.request("DELETE", uri, _kwargs=kwargs)

        return response

    def update(self, **kwargs) -> None:
        """
        Updates the space holiday. Attributes are updated in place with new information
        retrieved from the API.

        :calls: "PUT /space/{spaceId}/holidays/{holidayId}" \
		<https://fabman.io/api/v1/documentation#/spaces/putSpacesIdHolidaysHolidayid>
  
        :return: None -- attributes are updated in place
        :rtype: None
        """

        uri = f"/spaces/{self.space_id}/holidays/{self.id}"

        kwargs.update({"lockVersion": self.lockVersion})

        response = self._requester.request("PUT", uri, _kwargs=kwargs)

        data = response.json()

        for attr, val in data.items():
            setattr(self, attr, val)


class SpaceOpeningHours(FabmanObject):
    """Class for holding Opening Hours requests"""

    def __str__(self):
        out = f"Space #{self.space_id} Opening Hours:\n"
        out += "+-----------------+-----------------+\n"
        for day in self.days:
            out += f"{day['dayOfWeek']}: {day['fromTime']} - {day['untilTime']}\n"

        return out


class Space(FabmanObject):
    """Class for interacting with the Space endpoint on the Fabman API"""

    def __str__(self):
        return f"Space #{self.id}: {self.name}"

    def create_holiday(self, **kwargs) -> SpaceHoliday:
        """
        Creates a new holiday for the space.

        :calls: "POST /spaces/{spaceId}/holidays" \
		<https://fabman.io/api/v1/documentation#/spaces/postSpacesIdHolidays>
  
        :return: Object with new holiday information
        :rtype: fabman.space.SpaceHoliday
        """

        uri = f"/spaces/{self.id}/holidays"

        response = self._requester.request("POST", uri, _kwargs=kwargs)

        return SpaceHoliday(self._requester, response.json())

    def delete(self, **kwargs) -> requests.Response:
        """
        Deletes a space. *WARNING: This is irreversible.*

        :calls: "DELETE /spaces/{spaceId}" \
		<https://fabman.io/api/v1/documentation#/spaces/deleteSpacesId>

        :return: response information of call
        :rtype: requests.Response
        """

        uri = f"/spaces/{self.id}"

        response = self._requester.request("DELETE", uri, _kwargs=kwargs)

        return response

    def delete_calendar_token(self, **kwargs) -> requests.Response:
        """
        Deletes a space calendar token. *WARNING: This is irreversible.*

        :calls: "DELETE /spaces/{spaceId}/calendar-token" \
		<https://fabman.io/api/v1/documentation#/spaces/deleteSpacesIdCalendartoken>
  
        :return: response information of call
        :rtype: requests.Response
        """

        uri = f"/spaces/{self.id}/calendar-token"

        response = self._requester.request("DELETE", uri, _kwargs=kwargs)

        return response

    def get_billing_settings(self, **kwargs) -> SpaceBillingSettings:
        """
        Returns the billing settings for the space.

        :calls: "GET /spaces/{spaceId}/billing-settings" \
		<https://fabman.io/api/v1/documentation#/spaces/getSpacesIdBillingsettings>
  
        :return: Object with billing settings information
        :rtype: fabman.space.SpaceBillingSettings
        """
        if "billingSettings" in self._embedded:
            settings = self._embedded["billingSettings"]
            settings.update({"space_id": self.id})
            return SpaceBillingSettings(self._requester, settings)

        uri = f"/spaces/{self.id}/billing-settings"

        response = self._requester.request("GET", uri, _kwargs=kwargs)

        data = response.json()
        data.update({"space_id": self.id})

        return SpaceBillingSettings(self._requester, data)

    def get_holiday(self, holiday_id, **kwargs) -> SpaceHoliday:
        """
        Returns a specific holiday for the space given a holiday_id.

        :calls: "GET /spaces/{spaceId}/holidays/{holidayId}" \
		<https://fabman.io/api/v1/documentation#/spaces/getSpacesIdHolidaysHolidayid>
        
        :param holiday_id: ID of the holiday to retrieve
        :type holiday_id: int
        :return: Object with holiday information
        :rtype: fabman.space.SpaceHoliday
        """

        uri = f"/spaces/{self.id}/holidays/{holiday_id}"

        response = self._requester.request("GET", uri, _kwargs=kwargs)

        data = response.json()
        data.update({"space_id": self.id})

        return SpaceHoliday(self._requester, data)

    def get_holidays(self, **kwargs) -> Union[EmbeddedList, PaginatedList]:
        """
        Returns a list of holidays for the space. If the information is cached from an embedded call, 
        a list of SpaceHoliday objects is returned. Otherwise, a PaginatedList will be returned.

        :calls: "GET /spaces/{spaceId}/holidays" \
		<https://fabman.io/api/v1/documentation#/spaces/getSpacesIdHolidays>
  
        :return: List of SpaceHoliday Objects
        :rtype: list[fabman.space.SpaceHoliday] or fabman.paginated_list.PaginatedList
        """
        if "holidays" in self._embedded:
            out = []
            return EmbeddedList(
                SpaceHoliday,
                self._embedded["holidays"],
                self._requester,
                "GET",
                f"/spaces/{self.id}/holidays",
                extra_attribs={"space_id": self.id},
                kwargs=kwargs,
            )

        return PaginatedList(
            SpaceHoliday,
            self._requester,
            "GET",
            f"/spaces/{self.id}/holidays",
            extra_attribs={"space_id": self.id},
            kwargs=kwargs,
        )

    def get_opening_hours(self, **kwargs) -> SpaceOpeningHours:
        """
        Get the opening hours for the space.

        :calls: "GET /spaces/{spaceId}/opening-hours" \
		<https://fabman.io/api/v1/documentation#/spaces/getSpacesIdOpeninghours>

        :return: List of opening hours
        :rtype: list
        """

        if "openingHours" in self._embedded:
            data = {"days": self._embedded["openingHours"]}
        else:
            uri = f"/spaces/{self.id}/opening-hours"

            response = self._requester.request("GET", uri, _kwargs=kwargs)
            data = {"days": response.json()}

        data.update({"space_id": self.id})
        return SpaceOpeningHours(self._requester, data)

    def update(self, **kwargs) -> None:
        """
        Updates the space. Attributes are updated in place with new information
        from the API.

        :calls: "PUT /spaces/{spaceId}" \
		<https://fabman.io/api/v1/documentation#/spaces/putSpacesId>

        :return: None -- attributes are updated in place
        :rtype: None
        """

        uri = f"/spaces/{self.id}"

        kwargs.update({"lockVersion": self.lockVersion})

        response = self._requester.request("PUT", uri, _kwargs=kwargs)

        data = response.json()

        for attr, val in data.items():
            setattr(self, attr, val)

    def update_calendar_token(self, **kwargs) -> requests.Response:
        """
        Updates the space calendar token and returns a link to the new calendar ics
        file. Calendar Token attribute is updated in place with new information.

        :calls: "PUT /spaces/{spaceId}/calendar-token" \
		<https://fabman.io/api/v1/documentation#/spaces/putSpacesIdCalendartoken>

        :return: response information of call
        :rtype: requests.Response
        """

        uri = f"/spaces/{self.id}/calendar-token"

        response = self._requester.request("PUT", uri, _kwargs=kwargs)

        data = response.json()

        token = data["calendarUrl"].split("/")[-1].split(".")[0]
        setattr(self, "calendarToken", token)
        setattr(self, "calendarUrl", data["calendarUrl"])

        return response

    def update_opening_hours(self, **kwargs) -> SpaceOpeningHours:
        """
        Updates the opening hours of a space. If the the openingHours key is
        present in the _embedded attribute, the opening hours are updated in
        place with information returned by the api.

        :calls: "PUT /spaces/{spaceId}/opening-hours" \
		<https://fabman.io/api/v1/documentation#/spaces/putSpacesIdOpeninghours>
  
        :return: response information of call
        :rtype: SpaceOpeningHours
        """

        uri = f"/spaces/{self.id}/opening-hours"

        response = self._requester.request("PUT", uri, _kwargs=kwargs)

        data = response.json()
        out = {"days": data}
        out.update({"space_id": self.id})

        return SpaceOpeningHours(self._requester, out)



"""Tests for the Space class."""
# pylint: disable=missing-docstring, invalid-name, unused-argument, protected-access

import unittest

import requests
import requests_mock

from fabman import Fabman
from fabman.embedded_list import EmbeddedList
from fabman.paginated_list import PaginatedList
from fabman.space import Space, SpaceBillingSettings, SpaceHoliday, SpaceOpeningHours
from tests import settings
from tests.util import register_uris, validate_update


@requests_mock.Mocker()
class TestSpace(unittest.TestCase):
    def setUp(self):
        self.fabman = Fabman(settings.API_KEY)

        with requests_mock.Mocker() as m:
            register_uris({"fabman": ["get_space_by_id"]}, m)

            self.space: Space = self.fabman.get_space(1)

    def test_instance(self, m):
        self.assertIsInstance(self.space, Space)

    def test_str(self, m):
        string = str(self.space)
        self.assertTrue(string.startswith("Space #1: Operations Center"))

    def test_create_holiday(self, m):
        register_uris({"space": ["create_holiday"]}, m)

        holiday = self.space.create_holiday(
            title="Test Holiday",
            fromDateTime="2023-06-28T00:00",
            untilDateTime="2023-06-28T23:59",
            affects="all",
        )

        self.assertIsInstance(holiday, SpaceHoliday)
        self.assertTrue(hasattr(holiday, "id"))
        self.assertTrue(holiday.id == 1)

    def test_delete(self, m):
        register_uris({"space": ["delete"]}, m)

        response = self.space.delete()
        self.assertTrue(m.called)
        self.assertIsInstance(response, requests.Response)
        self.assertTrue(response.status_code == 204)

    def test_delete_calendar_token(self, m):
        register_uris({"space": ["delete_calendar_token"]}, m)

        response = self.space.delete_calendar_token()
        self.assertTrue(m.called)
        self.assertIsInstance(response, requests.Response)
        self.assertTrue(response.status_code == 204)

    def test_get_billing_settings(self, m):
        register_uris({"space": ["get_billing_settings"]}, m)

        billing = self.space.get_billing_settings()
        self.assertIsInstance(billing, SpaceBillingSettings)
        self.assertTrue(hasattr(billing, "space_id"))

    def test_get_embedded(self, m):
        register_uris({"space": ["get_embedded"]}, m)

        space = self.fabman.get_space(
            1, embed=["billingSettings", "holidays", "openingHours"]
        )
        self.assertIsInstance(space, Space)
        self.assertTrue(hasattr(space, "_embedded"))

        # billingSettings
        self.assertTrue("billingSettings" in space._embedded)
        billing = space.get_billing_settings()
        self.assertIsInstance(billing, SpaceBillingSettings)
        self.assertTrue(hasattr(billing, "space_id"))

        # holidays
        self.assertTrue("holidays" in space._embedded)
        holidays = space.get_holidays()
        self.assertIsInstance(holidays, EmbeddedList)
        self.assertIsInstance(holidays[0], SpaceHoliday)
        self.assertTrue(hasattr(holidays[0], "space_id"))

        # openingHours
        self.assertTrue("openingHours" in space._embedded)
        opening_hours = space.get_opening_hours()
        self.assertIsInstance(opening_hours, SpaceOpeningHours)
        self.assertEqual(opening_hours.space_id, 1)

    def test_get_holiday(self, m):
        register_uris({"space": ["get_holiday"]}, m)

        holiday = self.space.get_holiday(1)
        self.assertIsInstance(holiday, SpaceHoliday)
        self.assertTrue(hasattr(holiday, "id"))
        self.assertTrue(holiday.id == 1)

    def test_get_holidays(self, m):
        register_uris({"space": ["get_holidays"]}, m)

        holiday = self.space.get_holidays()
        self.assertIsInstance(holiday, PaginatedList)
        self.assertIsInstance(holiday[0], SpaceHoliday)
        self.assertTrue(hasattr(holiday[0], "id"))
        self.assertTrue(holiday[0].id == 1)

    def test_get_opening_hours(self, m):
        register_uris({"space": ["get_opening_hours"]}, m)

        resp = self.space.get_opening_hours()
        self.assertTrue(m.called)
        self.assertIsInstance(resp, SpaceOpeningHours)
        self.assertTrue(resp.space_id == 1)

        self.assertTrue("Space #1 Opening Hours" in str(resp))

    def test_update_calendar_token(self, m):
        register_uris({"space": ["update_calendar_token"]}, m)

        resp = self.space.update_calendar_token()
        self.assertTrue(m.called)
        self.assertIsInstance(resp, requests.Response)
        self.assertTrue(resp.status_code == 200)
        self.assertTrue(hasattr(self.space, "calendarToken"))
        self.assertTrue(self.space.calendarToken == "abcdef-1234-5678")

    def test_update_hours(self, m):
        register_uris({"space": ["update_opening_hours"]}, m)

        resp = self.space.update_opening_hours()
        self.assertTrue(m.called)
        self.assertIsInstance(resp, SpaceOpeningHours)
        self.assertTrue(resp.space_id == 1)

    def test_update(self, m):
        m.register_uri(
            "PUT",
            f"{settings.BASE_URL_WITH_VERSION}/spaces/1",
            text=validate_update,
            status_code=200,
        )

        self.space.update(name="Operations Center")

        self.assertTrue(m.called)


@requests_mock.Mocker()
class testSpaceBillingSettings(unittest.TestCase):
    def setUp(self):
        self.fabman = Fabman(settings.API_KEY)

        with requests_mock.Mocker() as m:
            register_uris(
                {"space": ["get_billing_settings"], "fabman": ["get_space_by_id"]}, m
            )

            self.space: Space = self.fabman.get_space(1)
            self.billing: SpaceBillingSettings = self.space.get_billing_settings()

    def test_instance(self, m):
        self.assertIsInstance(self.billing, SpaceBillingSettings)

    def test_str(self, m):
        string = str(self.billing)
        self.assertTrue(string.startswith("SpaceBillingSettings for space #1"))

    def test_delete_stripe(self, m):
        register_uris({"space": ["delete_stripe"]}, m)

        resp = self.billing.delete_stripe()
        self.assertTrue(m.called)
        self.assertIsInstance(resp, requests.Response)
        self.assertTrue(resp.status_code == 204)

    def test_update(self, m):
        m.register_uri(
            "PUT",
            f"{settings.BASE_URL_WITH_VERSION}/spaces/1/billing-settings",
            text=validate_update,
            status_code=200,
        )

        self.billing.update()
        self.assertTrue(m.called)


@requests_mock.Mocker()
class TestSpaceHoliday(unittest.TestCase):
    def setUp(self):
        self.fabman = Fabman(settings.API_KEY)

        with requests_mock.Mocker() as m:
            register_uris({"space": ["get_holiday"], "fabman": ["get_space_by_id"]}, m)

            self.space: Space = self.fabman.get_space(1)
            self.holiday: SpaceHoliday = self.space.get_holiday(1)

    def test_instance(self, m):
        self.assertIsInstance(self.holiday, SpaceHoliday)

    def test_str(self, m):
        string = str(self.holiday)
        self.assertTrue(string.startswith("SpaceHoliday #1: Emissary's Day"))

    def test_delete(self, m):
        register_uris({"space": ["delete_holiday"]}, m)

        resp = self.holiday.delete()
        self.assertTrue(m.called)
        self.assertIsInstance(resp, requests.Response)
        self.assertTrue(resp.status_code == 204)

    def test_update(self, m):
        m.register_uri(
            "PUT",
            f"{settings.BASE_URL_WITH_VERSION}/spaces/1/holidays/1",
            text=validate_update,
            status_code=200,
        )

        self.holiday.update()
        self.assertTrue(m.called)



"""Defines the ApiKey class"""

import requests

from fabman.fabman_object import FabmanObject


class ApiKeyToken(FabmanObject):
    """Simple Class to hold the token for an api-key"""

    def __str__(self):
        return f"ApiKeyToken #{self.api_key_id}: {self.token}"


class ApiKey(FabmanObject):
    """
    Class for interacting with the api-keys endpoints on the Fabman API
    """

    def __str__(self):
        return f"ApiKey #{self.id}: {self.label}"

    def delete(self, **kwargs) -> requests.Response:
        """
        Deletes an api-key. *WARNING: This is irreversible.*

        :calls: "DELETE /api-keys/{apiKeyId}" \
		<https://fabman.io/api/v1/documentation#/api-keys/deleteApiKeysId>
  
        :returns: Response from the API with status code 204 (No Content) if successful
        :rtype: requests.Response
        """

        uri = f"/api-keys/{self.id}"

        response = self._requester.request("DELETE", uri, _kwargs=kwargs)

        return response

    def get_token(self, **kwargs) -> ApiKeyToken:
        """
        Returns the token for an api-key. Must be called with a valid api-key.

        :calls: "GET /api-keys/{apiKeyId}/token" \
		<https://fabman.io/api/v1/documentation#/api-keys/getApiKeysIdToken>
  
        :returns: Response from the API with status code 200 (OK) if successful
        :rtype: requests.Response
        """

        uri = f"/api-keys/{self.id}/token"

        response = self._requester.request("GET", uri, _kwargs=kwargs)
        data = response.json()

        data.update({"api_key_id": self.id})

        return ApiKeyToken(self._requester, data)

    def update(self, **kwargs) -> None:
        """
        Updates the api-key. Attributes are updated in place with new information
        returned by the API.

        :calls: "PUT /api-keys/{apiKeyId}" \
		<https://fabman.io/api/v1/documentation#/api-keys/putApiKeysId>
  
        :returns: None
        :rtype: None
        """

        uri = f"/api-keys/{self.id}"

        kwargs.update({"lockVersion": self.lockVersion})
        response = self._requester.request("PUT", uri, _kwargs=kwargs)

        data = response.json()

        for attr, val in data.items():
            setattr(self, attr, val)



#!/usr/bin/python3
"""Main file for the Fabman API library.
"""

import warnings

import requests

from fabman.account import Account
from fabman.api_key import ApiKey
from fabman.booking import Booking
from fabman.charge import Charge
from fabman.invoice import Invoice
from fabman.job import Job
from fabman.member import Member
from fabman.package import Package
from fabman.paginated_list import PaginatedList
from fabman.payment import Payment
from fabman.requester import Requester
from fabman.resource import Resource
from fabman.resource_log import ResourceLog
from fabman.resource_type import ResourceType
from fabman.space import Space
from fabman.training_course import TrainingCourse
from fabman.webhook import Webhook


class Fabman(object):
    """
    The main class to be instantiated to provide access to the Fabman api.
    """

    def __init__(self, access_token: str, base_url="https://fabman.io/api/v1"):
        """
        Initializes the Fabman class with the given access token and base url.
        All methods take kwargs as their arguments, please refer to the Fabman API
        for more information

        :param access_token: The access token to use for the API
        :type access_token: str
        :param base_url (optional): The base url to use for the API
        :type base_url: str
        """

        if "https://" not in base_url:
            warnings.warn(
                "Please use HTTPS when possible. Fabman API may not respond as intended and user"
                "data will not be secure",
                UserWarning,
            )

        if "://" not in base_url:
            warnings.warn(
                "An invalid `bad_url` provided. Will likely not work as intended.",
                UserWarning,
            )
        if not access_token or access_token == "":
            raise ValueError("No access token provided")

        # sanitize access token and base url
        access_token = access_token.strip()
        if base_url[-1] == "/":
            base_url = base_url[:-1]

        self.__requester = Requester(base_url, access_token)

    def create_api_key(self, **kwargs) -> ApiKey:
        """
        Creates a new API key for a member.

        :calls: "POST /api-keys" \
        <https://fabman.io/api/v1/documentation#/api-keys/postApikeys>
        
        :returns: :code:`ApiKey` object if successful
        :rtype: :code:`fabman.ApiKey`
        """

        uri = "/api-keys"
        response = self.__requester.request("POST", uri, _kwargs=kwargs)

        return ApiKey(self.__requester, response.json())

    def create_booking(self, **kwargs) -> Booking:
        """
        Creates a new booking in the Fabman database.

        :calls: "POST /bookings" \
		<https://fabman.io/api/v1/documentation#/bookings/postBookings>
  
        :returns: :code:`Booking` object if successful
        :rtype: :code:`fabman.Booking`
        """

        uri = "/bookings"

        response = self.__requester.request("POST", uri, _kwargs=kwargs)

        return Booking(self.__requester, response.json())

    def create_charge(self, **kwargs) -> Charge:
        """
        Creates a new charge in the Fabman database.

        :calls: "POST /charges" \
		<https://fabman.io/api/v1/documentation#/charges/postCharges>
  
        :returns: :code:`Charge` object if successful
        :rtype: :code:`fabman.Charge`
        """

        uri = "/charges"

        response = self.__requester.request("POST", uri, _kwargs=kwargs)

        return Charge(self.__requester, response.json())

    def create_invoice(self, **kwargs) -> Invoice:
        """
        Creates a new invoice in the Fabman database.

        :calls: "POST /invoices" \
		<https://fabman.io/api/v1/documentation#/invoices/postInvoices>
  
        :returns: :code:`Invoice` object if successful
        :rtype: :code:`fabman.Invoice`
        """

        uri = "/invoices"

        response = self.__requester.request("POST", uri, _kwargs=kwargs)

        return Invoice(self.__requester, response.json())

    def create_key_assignment(self, **kwargs) -> requests.Response:
        """
        Creates a new keycard assignment in the Fabman Database. Note: A better
        method to work with key assignments is to use the Member class and the
        associated MemberKey class that works with the key assignments endpoint.

        :calls: "POST /key-assignments" \
		<https://fabman.io/api/v1/documentation#/key-assignments/postKeyassignments>
  
        :returns: :code:`requests.Response` object if successful
        :rtype: :code:`requests.Response`
        """

        uri = "/key-assignments"

        response = self.__requester.request("POST", uri, _kwargs=kwargs)

        return response

    def create_member(self, **kwargs) -> Member:
        """
        Creates a new member in the Fabman database.

        :calls: "POST /members" \
        <https://fabman.io/api/v1/documentation#/members/postMembers>

        :returns: :code:`Member` object if successful
        :rtype: :code:`fabman.Member`

        """
        uri = "/members"

        response = self.__requester.request("POST", uri, _kwargs=kwargs)

        return Member(self.__requester, response.json())

    def create_package(self, **kwargs) -> Package:
        """
        Creates a new package in the Fabman database.

        :calls: "POST /packages" \
        <https://fabman.io/api/v1/documentation#/packages/postPackages>
        
        :returns: :code:`Package` object if successful
        :rtype: :code:`fabman.Package`
        """

        uri = "/packages"

        response = self.__requester.request("POST", uri, _kwargs=kwargs)

        return Package(self.__requester, response.json())

    def create_payment(self, **kwargs) -> Payment:
        """
        Creates a new payment in the Fabman database.

        :calls: "POST /payments" \
		<https://fabman.io/api/v1/documentation#/payments/postPayments>
  
        :returns: :code:`Payment` object if successful
        :rtype: :code:`fabman.Payment`
        """

        uri = "/payments"

        response = self.__requester.request("POST", uri, _kwargs=kwargs)

        return Payment(self.__requester, response.json())

    def create_resource(self, **kwargs) -> Resource:
        """
        Creates a new resource in the Fabman database.

        :calls: "POST /resources" \
		<https://fabman.io/api/v1/documentation#/resources/postResources>
  
        :returns: :code:`Resource` object if successful
        :rtype: :code:`fabman.Resource`
        """

        uri = "/resources"

        response = self.__requester.request("POST", uri, _kwargs=kwargs)

        return Resource(self.__requester, response.json())

    def create_resource_log(self, **kwargs) -> ResourceLog:
        """
        Creates a new resource log in the Fabman database. Use with caution.

        :calls: "POST /resource-logs" \
		<https://fabman.io/api/v1/documentation#/resource-logs/postResourcelogs>
  
        :returns: :code:`ResourceLog` object if successful
        :rtype: :code:`fabman.ResourceLog`
        """
        uri = "/resource-logs"

        response = self.__requester.request("POST", uri, _kwargs=kwargs)

        return ResourceLog(self.__requester, response.json())

    def create_resource_type(self, **kwargs) -> ResourceType:
        """
        Creates a new resource type in the Fabman database.

        :calls: "POST /resource-types" \
		<https://fabman.io/api/v1/documentation#/resource-types/postResourcetypes>
  
        :returns: :code:`ResourceType` object if successful
        :rtype: :code:`fabman.ResourceType`
        """
        uri = "/resource-types"

        response = self.__requester.request("POST", uri, _kwargs=kwargs)

        return ResourceType(self.__requester, response.json())

    def create_space(self, **kwargs) -> Space:
        """
        Creates a new space in the Fabman database.

        :calls: "POST /spaces" \
		<https://fabman.io/api/v1/documentation#/spaces/postSpaces>
  
        :returns: :code:`Space` object if successful
        :rtype: :code:`fabman.Space`
        """
        uri = "/spaces"

        response = self.__requester.request("POST", uri, _kwargs=kwargs)

        return Space(self.__requester, response.json())

    def create_training_course(self, **kwargs) -> TrainingCourse:
        """
        Creates a new Training course in the Fabman database.

        :calls: "POST /training-courses" \
        <https://fabman.io/api/v1/documentation#/training-courses/postTrainingcourses>
        
        :returns: :code:`TrainingCourse` object if successful
        :rtype: :code:`fabman.TrainingCourse`
        """

        uri = "/training-courses"

        response = self.__requester.request("POST", uri, _kwargs=kwargs)

        return TrainingCourse(self.__requester, response.json())

    def create_webhook(self, **kwargs) -> Webhook:
        """
        Creates a new webhook in the Fabman database.

        :calls: "POST /webhooks" \
        <https://fabman.io/api/v1/documentation#/webhooks/postWebhooks>
        
        :returns: :code:`Webhook` object if successful
        :rtype: :code:`fabman.Webhook`
        """

        uri = "/webhooks"

        response = self.__requester.request("POST", uri, _kwargs=kwargs)

        return Webhook(self.__requester, response.json())

    def get_account(self, account_id, **kwargs) -> Account:
        """
        Get a single account by its ID. Note: for most users, the only account
        retrievable is the account of the API key holder. Majority of endpoints
        are not implemented as they require superuser privileges.

        :calls: "GET /accounts/{account_id}" \
        <https://fabman.io/api/v1/documentation#/accounts/getAccountsId>
        
        :param account_id: The id of the account to retrieve
        :type account_id: int
        :returns: :code:`Account` object if successful
        :rtype: :code:`fabman.Account`
        """

        uri = f"/accounts/{account_id}"

        response = self.__requester.request("GET", uri, _kwargs=kwargs)

        return Account(self.__requester, response.json())

    def get_accounts(self, **kwargs) -> PaginatedList:
        """
        Get a list of accounts. Note, for most users this will only return the account
        of the API key holder. Most documented endpoints are unimplemented as a result.

        :calls: "GET /accounts" \
        <https://fabman.io/api/v1/documentation#/accounts/getAccounts>
        
        :returns: :code:`PaginatedList` object if successful
        :rtype: :code:`fabman.PaginatedList`
        """
        return PaginatedList(Account, self.__requester, "GET", "/accounts", **kwargs)

    def get_api_key(self, key_id, **kwargs) -> ApiKey:
        """
        Get information about a single API key by its ID.

        :calls: "GET /api-keys/{key_id}" \
        <https://fabman.io/api/v1/documentation#/api-keys/getApikeysId>
        
        :param key_id: The id of the API key to retrieve
        :type key_id: int
        :returns: :code:`ApiKey` object if successful
        :rtype: :code:`fabman.ApiKey`
        """

        uri = f"/api-keys/{key_id}"

        response = self.__requester.request("GET", uri, _kwargs=kwargs)

        return ApiKey(self.__requester, response.json())

    def get_api_keys(self, **kwargs) -> PaginatedList:
        """
        Get a list of API keys. Can specify filters, search string, etc.

        :calls: "GET /api-keys" \
		<https://fabman.io/api/v1/documentation#/api-keys/getApikeys>
  
        :returns: :code:`PaginatedList` object if successful
        :rtype: :code:`fabman.PaginatedList`
        """

        return PaginatedList(ApiKey, self.__requester, "GET", "/api-keys", **kwargs)

    def get_booking(self, booking_id, **kwargs) -> Booking:
        """
        Get a single booking by its ID.

        :calls: "GET /bookings/{id}" \
		<https://fabman.io/api/v1/documentation#/bookings/getBookingsId>

        :param booking_id: The id of the booking to retrieve
        :type booking_id: int
        :returns: :code:`Booking` object if successful
        :rtype: :code:`fabman.Booking`
        """
        uri = f"/bookings/{booking_id}"

        response = self.__requester.request("GET", uri, _kwargs=kwargs)

        return Booking(self.__requester, response.json())

    def get_bookings(self, **kwargs) -> PaginatedList:
        """
        Retrieves a PaginatedList of bookings from the API. Can specify filters, search
        string, etc.

        :calls: "GET /bookings" \
		<https://fabman.io/api/v1/documentation#/bookings/getBookings>
  
        :returns: :code:`PaginatedList` object if successful
        :rtype: :code:`fabman.PaginatedList`
        """
        return PaginatedList(Booking, self.__requester, "GET", "/bookings", **kwargs)

    def get_charge(self, charge_id: int, **kwargs) -> Charge:
        """
        Retrieves a single Charge given the charge_id.

        :calls: "Get /charges/{id}" \
		<https://fabman.io/api/v1/documentation#/charges/getChargesId>

        :param charge_id: The id of the charge to retrieve
        :type charge_id: int
        :returns: :code:`Charge` object if successful
        :rtype: :code:`fabman.Charge`
        """

        uri = f"/charges/{charge_id}"

        response = self.__requester.request("GET", uri, _kwargs=kwargs)

        return Charge(self.__requester, response.json())

    def get_charges(self, **kwargs) -> PaginatedList:
        """
        Retrieves all charges from the API. Can specify filters, search string, etc.

        :calls: "GET /charges" \
		<https://fabman.io/api/v1/documentation#/charges/getCharges>
  
        :returns: :code:`PaginatedList` object if successful
        :rtype: :code:`fabman.PaginatedList`
        """

        return PaginatedList(Charge, self.__requester, "GET", "/charges", **kwargs)

    def get_invoice(self, invoice_id, **kwargs) -> Invoice:
        """
        Retrieve information of a single invoice by its ID.

        :calls: "GET /invoices/{id}" \
		<https://fabman.io/api/v1/documentation#/invoices/getInvoicesId>
  
        :param invoice_id: The id of the invoice to retrieve
        :type invoice_id: int
        :returns: :code:`Invoice` object if successful
        :rtype: :code:`fabman.Invoice`
        """

        uri = f"/invoices/{invoice_id}"

        response = self.__requester.request("GET", uri, _kwargs=kwargs)

        return Invoice(self.__requester, response.json())

    def get_invoices(self, **kwargs) -> PaginatedList:
        """
        Retrieves list of all invoices. Can specify filters, search string, etc.

        :calls: "GET /invoices" \
		<https://fabman.io/api/v1/documentation#/invoices/getInvoices>
  
        :returns: :code:`PaginatedList` object if successful
        :rtype: :code:`fabman.PaginatedList`
        """

        return PaginatedList(Invoice, self.__requester, "GET", "/invoices", **kwargs)

    def get_job(self, job_id, **kwargs) -> Job:
        """
        Retrieve a single job from the Fabman API given the job_id.

        :calls: "GET /jobs/{id}" \
		<https://fabman.io/api/v1/documentation#/jobs/getJobsId>
  
        :param job_id: The id of the job to retrieve
        :type job_id: int
        :returns: :code:`Job` object if successful
        :rtype: :code:`Job`
        """
        uri = f"/jobs/{job_id}"
        response = self.__requester.request("GET", uri, _kwargs=kwargs)

        return Job(self.__requester, response.json())

    def get_jobs(self, **kwargs) -> PaginatedList:
        """
        Retrieves a list of all jobs. Can specify filters, search string, etc.

        :calls: "GET /jobs" \
		<https://fabman.io/api/v1/documentation#/jobs/getJobs>
  
        :returns: :code:`PaginatedList` object if successful
        :rtype: :code:`fabman.PaginatedList`
        """

        return PaginatedList(Job, self.__requester, "GET", "/jobs", **kwargs)

    def get_member(self, member_id: int, **kwargs):
        """Retrieves a member from the API give their id
        :calls: "GET /members/{id}" \
		<https://fabman.io/api/v1/documentation#/members/getMembersId>

        :param member_id: The id of the member to retrieve
        :type member_id: int
        :returns: :code:`Member` object if successful
        """
        uri = f"/members/{member_id}"

        response = self.__requester.request("GET", uri, _kwargs=kwargs)

        return Member(self.__requester, response.json())

    def get_members(self, **kwargs):
        """Get all of the members in the Fabman database. Can specify filters,
        search string, result limits, offsets, and sorting. Refer to the Fabman API
        documentation.

        :calls: "GET /members" \
		<https://fabman.io/api/v1/documentation#/members/getMembers>

        :returns: :code:`PaginatedList` object if successful
        :rtype: :code:`fabman.PaginatedList`
        """

        return PaginatedList(Member, self.__requester, "GET", "/members", **kwargs)

    def get_package(self, package_id: int, **kwargs) -> Package:
        """
        Retrieves a single Package given a package_id

        :calls: "GET /packages/{id}" \
		<https://fabman.io/api/v1/documentation#/packages/getPackagesId>
  
        :param package_id: The id of the package to retrieve
        :type package_id: int
        :returns: :code:`Package` object if successful
        :rtype: :code:`fabman.Package`
        """

        uri = f"/packages/{package_id}"

        response = self.__requester.request("GET", uri, _kwargs=kwargs)

        return Package(self.__requester, response.json())

    def get_packages(self, **kwargs) -> PaginatedList:
        """
        Retrieves a list of packages. Can specify filters, search string, etc.

        :calls: "GET /packages" \
		<https://fabman.io/api/v1/documentation#/packages/getPackages>
  
        :returns: :code:`PaginatedList` object if successful
        :rtype: :code:`fabman.PaginatedList`
        """

        return PaginatedList(Package, self.__requester, "GET", "/packages", **kwargs)

    def get_payment(self, payment_id, **kwargs) -> Payment:
        """
        Retrieves a single payment given a payment_id.

        :calls: "GET /payments/{id}" \
		<https://fabman.io/api/v1/documentation#/payments/getPaymentsId>

        :param payment_id: The id of the payment to retrieve
        :type payment_id: int
        :returns: :code:`Payment` object if successful
        :rtype: :code:`fabman.Payment`
        """

        uri = f"/payments/{payment_id}"

        response = self.__requester.request("GET", uri, _kwargs=kwargs)

        return Payment(self.__requester, response.json())

    def get_payments(self, **kwargs) -> PaginatedList:
        """
        Retrieves a list of payments. Can specify filters, search string, etc.

        :calls: "GET /payments" \
		<https://fabman.io/api/v1/documentation#/payments/getPayments>

        :returns: :code:`PaginatedList` object if successful
        :rtype: :code:`fabman.PaginatedList`
        """

        return PaginatedList(Payment, self.__requester, "GET", "/payments", **kwargs)

    def get_resource(self, resource_id: int, **kwargs):
        """
        Get a single resource by its ID. Embed information is also available.

        :calls: "GET /resources/{id}" \
		<https://fabman.io/api/v1/documentation#/resources/getResourcesId>

        :param resource_id: The id of the resource to retrieve
        :type resource_id: int
        :returns: :code:`Resource` object if successful
        :rtype: :code:`fabman.Resource`
        """
        uri = f"/resources/{resource_id}"

        response = self.__requester.request("GET", uri, _kwargs=kwargs)

        return Resource(self.__requester, response.json())

    def get_resources(self, **kwargs) -> PaginatedList:
        """
        Get list of available resources (e.g. doors, printers, etc.) Limit, offset,
        and a number of filters are available. Refer to the appropriate documentation.

        :calls: "GET /resources" \
		<https://fabman.io/api/v1/documentation#/resources/getResources>
  
        :returns: :code:`PaginatedList` object if successful
        :rtype: :code:`fabman.PaginatedList`
        """
        return PaginatedList(
            Resource,
            self.__requester,
            "GET",
            "/resources",
            **kwargs,
        )

    def get_resource_log(self, resource_log_id, **kwargs) -> ResourceLog:
        """
        Retrieves a single Resource Log from the API.

        :calls: "GET /resource-logs/{id}" \
		<https://fabman.io/api/v1/documentation#/resource-logs/getResourcelogsId>

        :param resource_log_id: The id of the resource log to retrieve
        :type resource_log_id: int
        :returns: :code:`ResourceLog` object if successful
        :rtype: :code:`fabman.ResourceLog`
        """

        uri = f"/resource-logs/{resource_log_id}"

        response = self.__requester.request("GET", uri, _kwargs=kwargs)

        return ResourceLog(self.__requester, response.json())

    def get_resource_logs(self, **kwargs) -> PaginatedList:
        """
        Retrieves a list of resource logs. Can specify filters, search string, etc.

        :calls: "GET /resource-logs" \
		<https://fabman.io/api/v1/documentation#/resource-logs/getResourcelogs>
  
        :returns: :code:`PaginatedList` object if successful
        :rtype: :code:`fabman.PaginatedList`
        """

        return PaginatedList(
            ResourceLog, self.__requester, "GET", "/resource-logs", **kwargs
        )

    def get_resource_types(self, **kwargs) -> PaginatedList:
        """
        Retrieves a list of resource types. Can specify filters, search string, etc.

        :calls: "GET /resource-types" \
		<https://fabman.io/api/v1/documentation#/resource-types/getResourcetypes>
  
        :returns: :code:`PaginatedList` object if successful
        :rtype: :code:`fabman.PaginatedList`
        """

        return PaginatedList(
            ResourceType, self.__requester, "GET", "/resource-types", **kwargs
        )

    def get_space(self, space_id, **kwargs) -> Space:
        """
        Retrieves a single space given a space_id.

        :calls: "GET /spaces/{id}" \
		<https://fabman.io/api/v1/documentation#/spaces/getSpacesId>

        :param space_id: The id of the space to retrieve
        :type space_id: int
        :returns: :code:`Space` object if successful
        :rtype: :code:`fabman.Space`
        """

        uri = f"/spaces/{space_id}"

        response = self.__requester.request("GET", uri, _kwargs=kwargs)

        return Space(self.__requester, response.json())

    def get_spaces(self, **kwargs) -> PaginatedList:
        """
        Retrieves a list of Spaces. Can specify filters, search string, etc.

        :calls: "GET /spaces" \
		<https://fabman.io/api/v1/documentation#/spaces/getSpaces>
  
        :returns: :code:`PaginatedList` object if successful
        :rtype: :code:`fabman.PaginatedList`
        """

        return PaginatedList(Space, self.__requester, "GET", "/spaces", **kwargs)

    def get_training_course(self, course_id, **kwargs):
        """
        Retrieve a single TrainingCourse object from the API.

        :calls: "GET /training-courses/{id}" \
		<https://fabman.io/api/v1/documentation#/training-courses/getTrainingcoursesId>
  
        :param course_id: The id of the training course to retrieve
        :type course_id: int
        :returns: :code:`TrainingCourse` object if successful
        :rtype: :code:`fabman.TrainingCourse`
        """
        uri = f"/training-courses/{course_id}"

        response = self.__requester.request("GET", uri, _kwargs=kwargs)

        return TrainingCourse(self.__requester, response.json())

    def get_training_courses(self, **kwargs):
        """
        Retrieves a PaginatedList of Training Courses available on the api.

        :calls: "GET /training-courses" \
		<https://fabman.io/api/v1/documentation#/training-courses/getTrainingcourses>
  
        :returns: :code:`PaginatedList` object if successful
        :rtype: :code:`fabman.PaginatedList`
        """
        return PaginatedList(
            TrainingCourse, self.__requester, "GET", "/training-courses", **kwargs
        )

    def get_user(self, **kwargs) -> Member:
        """Gets authenticated user information from the API as Member object. Does
        not returns state and superuser information.
        :calls: "GET /user/me" \
		<https://fabman.io/api/v1/documentation#/user/getUserMe>

        :returns: :code:`Member` object if successful
        :rtype: :code:`fabman.Member`
        """
        uri = "/user/me"

        response = self.__requester.request("GET", uri, _kwargs=kwargs)

        return Member(self.__requester, response.json()["members"][0])

    def get_webhook(self, webhook_id, **kwargs) -> Webhook:
        """
        Retrieves a single webhook given a webhook_id.

        :calls: "GET /webhooks/{id}" \
		<https://fabman.io/api/v1/documentation#/webhooks/getWebhooksId>
  
        :param webhook_id: The id of the webhook to retrieve
        :type webhook_id: int
        :returns: :code:`Webhook` object if successful
        :rtype: :code:`fabman.Webhook`
        """

        uri = f"/webhooks/{webhook_id}"

        response = self.__requester.request("GET", uri, _kwargs=kwargs)

        return Webhook(self.__requester, response.json())

    def get_webhooks(self, **kwargs) -> PaginatedList:
        """
        Retrieves a list of webhooks. Can specify filters, search string, etc.

        :calls: "GET /webhooks" \
		<https://fabman.io/api/v1/documentation#/webhooks/getWebhooks>
  
        :returns: :code:`PaginatedList` object if successful
        :rtype: :code:`fabman.PaginatedList`
        """

        return PaginatedList(Webhook, self.__requester, "GET", "/webhooks", **kwargs)



"""The Requester module handles all requests to the API. This should not be used 
or called directly as it is meant to be used internally. All accesses should be
made directly through the Fabman class found in fabman/fabman.py
"""
import logging
import warnings
from pprint import pformat
from time import sleep
from typing import Optional

import requests

from fabman.exceptions import (
    BadRequest,
    Conflict,
    FabmanException,
    ForbiddenError,
    InvalidAccessToken,
    RateLimitExceeded,
    ResourceDoesNotExist,
    Unauthorized,
    UnprocessableEntity,
)
from fabman.util import clean_headers

logger = logging.getLogger(__name__)

CACHE_SIZE = 4


class Requester(object):
    """Main class responsible for handling all http requests to the API.
    Based on canvasapi.requester.Requester found at
    https://github.com/ucfopen/canvasapi/blob/develop/canvasapi/requester.py
    """

    def __init__(self, base_url: str, access_token: str) -> None:
        """
        :param base_url: The base URL of the Fabman instance's API.
        :type base_url: str
        :param access_token: The API key to authenticate requests with.
        :type access_token: str
        """

        self.base_url = base_url
        self.__access_token = access_token
        self.__session = requests.Session()
        self.__cache = []

    def _delete_request(
        self, url: str, headers: dict, data: Optional[dict] = None, **kwargs
    ) -> requests.Response:
        """
        Handles a delete request to the API. Should never be called directly
        """

        return self.__session.delete(url, headers=headers, data=data, **kwargs)

    def _get_request(
        self, url: str, headers: dict, params: Optional[dict] = None, **kwargs
    ) -> requests.Response:
        """
        Handles a get request to the API. Should never be called directly
        """

        return self.__session.get(url, headers=headers, params=params, **kwargs)

    def _post_request(
        self, url: str, headers: dict, data: Optional[dict] = None, **kwargs
    ) -> requests.Response:
        """
        Handles a post request to the API. Should never be called directly
        """

        return self.__session.post(url, headers=headers, data=data, **kwargs)

    def _put_request(
        self, url: str, headers: dict, data: Optional[dict] = None, **kwargs
    ) -> requests.Response:
        """
        Handles a put request to the API. Should never be called directly
        """

        return self.__session.put(url, headers=headers, data=data, **kwargs)

    def request(
        self,
        method: str,
        endpoint: Optional[str] = None,
        headers: Optional[dict] = None,
        use_auth: Optional[bool] = True,
        _url: Optional[str] = None,
        _kwargs: Optional[dict] = None,
        json: Optional[bool] = False,
        **kwargs,
    ) -> requests.Response:
        """
        Main method for handling requests to the API. Should never be called directly except for
        testing or from the Fabman class.

        :param method: The HTTP method to use for the request.
        :type method: str
        :param endpoint: The API endpoint to call.
        :type endpoint: str
        :param headers: HTTP headers to send with the request.
        :type headers: dict
        :param use_auth: Whether or not to include the access token in the request.
        :type use_auth: bool
        :param _url: The full URL to use for the request. This overrides the base_url and endpoint.
        :type _url: str
        :param _kwargs: Keyword arguments from the calling argument. These are morphed into the params or body of the request depending on :code:`method`.
        :type _kwargs: dict
        :param json: Whether or not to send the data as JSON.
        :type json: bool

        :return: The response object if the call was successful
        :rtype: requests.Response
        """
        full_url = _url if _url else f"{self.base_url}{endpoint}"

        if not headers:
            headers = {}

        if use_auth:
            auth_header = {"Authorization": f"Bearer {self.__access_token}"}
            headers.update(auth_header)

        # Add kwargs to _kwargs
        if _kwargs is None:
            _kwargs = kwargs
        else:
            _kwargs.update(kwargs)

        # Determine the appropriate request method.
        if method == "GET":
            req_method = self._get_request
        elif method == "POST":
            req_method = self._post_request
        elif method == "DELETE":
            req_method = self._delete_request
        elif method == "PUT":
            req_method = self._put_request
        else:
            raise ValueError(f"Invalid method {method}")

        logger.info("Request: %s %s", method, full_url)
        logger.debug(
            "Headers %s",
            pformat(clean_headers(headers), indent=2, width=80, compact=True),
        )

        response = req_method(full_url, headers, _kwargs, json=json)
        logger.info("Response: %s %s %s", method, full_url, response.status_code)
        logger.debug("Headers: %s", pformat(clean_headers(response.headers)))

        try:
            logger.debug("Data: %s", pformat(response.content.decode("utf-8")))
        except UnicodeDecodeError:
            logger.debug("Data: %s", pformat(response.content))
        except AttributeError:
            # Response has no content
            logger.debug("No data")

        # add response to cache
        if len(self.__cache) >= CACHE_SIZE:
            self.__cache.pop()
        self.__cache.insert(0, response)

        # Raise for status codes
        if response.status_code == 400:
            raise BadRequest(response.text)
        if response.status_code == 204:
            if method != "DELETE":
                warnings.warn(
                    "204 No Content returned. This likely means there is no information"
                    "at the resource.",
                    UserWarning,
                )
            return response
        if response.status_code == 401:
            if "WWW-Authenticate" in response.headers:
                raise InvalidAccessToken(response.json())
            raise Unauthorized(response.json())
        if response.status_code == 403:
            raise ForbiddenError(response.json())
        if response.status_code == 404:
            raise ResourceDoesNotExist("Not found")
        if response.status_code == 409:
            raise Conflict(response.text)
        if response.status_code == 422:
            raise UnprocessableEntity(response.text)
        # rate limit is number of request/second at time of writing. Unable to hit the
        # rate limit on home connection to verify this handling works. May break in the future.
        if response.status_code == 429:
            raise RateLimitExceeded(
                "Rate Limit Exceeded. Too many requests in a short amount of time. Retry in at least 2 seconds."
                "https://github.com/FabmanHQ/fabman-api#rate-limiting"
            )

        if response.status_code > 400:
            # catch all for other (bad)codes
            raise FabmanException(
                f"Encountered an error: status code {response.status_code}"
            )

        return response



"""Test for the ResourceType class"""
# pylint: disable=missing-docstring, invalid-name, unused-argument

import unittest

import requests
import requests_mock

from fabman import Fabman
from fabman.resource_type import ResourceType
from tests import settings
from tests.util import register_uris, validate_update


@requests_mock.Mocker()
class TestResourceType(unittest.TestCase):
    def setUp(self):
        self.fabman = Fabman(settings.API_KEY)

        with requests_mock.Mocker() as m:
            register_uris({"fabman": ["get_resource_types"]}, m)

            self.resource_type: ResourceType = self.fabman.get_resource_types()[0]

    def test_instance(self, m):
        self.assertIsInstance(self.resource_type, ResourceType)

    def test_str(self, m):
        string = str(self.resource_type)
        self.assertTrue(string == "ResourceType #1: Starship")

    def test_delete(self, m):
        register_uris({"resource_type": ["delete"]}, m)

        resp = self.resource_type.delete()
        self.assertTrue(m.called)
        self.assertIsInstance(resp, requests.Response)
        self.assertTrue(resp.status_code == 204)

    def test_update(self, m):
        register_uris({"resource_type": ["update"]}, m)

        self.resource_type.update(name="Victory Class Starship")
        self.assertTrue(m.called)
        self.assertTrue(self.resource_type.name == "Victory Class Starship")



"""Tests for the fabman module."""
# pylint: disable=missing-docstring, invalid-name, unused-argument, protected-access
import unittest
import warnings

import requests
import requests_mock

from fabman import Fabman
from fabman.account import Account
from fabman.api_key import ApiKey
from fabman.booking import Booking
from fabman.charge import Charge
from fabman.fabman_object import FabmanObject
from fabman.invoice import Invoice
from fabman.job import Job
from fabman.member import Member
from fabman.package import Package
from fabman.paginated_list import PaginatedList
from fabman.payment import Payment
from fabman.requester import Requester
from fabman.resource import Resource
from fabman.resource_log import ResourceLog
from fabman.resource_type import ResourceType
from fabman.space import Space
from fabman.training_course import TrainingCourse
from fabman.webhook import Webhook
from tests import settings
from tests.util import register_uris


@requests_mock.Mocker()
class TestFabman(unittest.TestCase):
    def setUp(self):
        self.fabman = Fabman(settings.API_KEY)

    # Test initializing the Fabman instance
    def test_init_no_api_key(self, m):
        with self.assertRaises(ValueError, msg="No access token provided"):
            Fabman("")

    def test_init_empty_api_key(self, m):
        with self.assertRaises(ValueError, msg="No access token provided"):
            Fabman("")

    def test_init_warnings_for_http(self, m):
        with warnings.catch_warnings(record=True):
            Fabman(settings.API_KEY, settings.BASE_URL_AS_HTTP)
            self.assertRaises(
                UserWarning,
                msg=(
                    "Please use HTTPS when possible. Fabman API may not respond as intended and"
                    "user data will not be secure",
                ),
            )

    def test_init_warnings_for_bad_url(self, m):
        with warnings.catch_warnings(record=True):
            Fabman(settings.API_KEY, settings.BASE_URL_AS_INVALID)
            self.assertRaises(
                UserWarning,
                msg=(
                    "An invalid `bad_url` provided. Will likely not work as intended."
                ),
            )

    def test_init_trailing_slash(self, m):
        register_uris({"fabman": ["get_member_by_id"]}, m)
        fabman = Fabman(settings.API_KEY, settings.BASE_URL_WITH_TRAILING_SLASH)
        member = fabman.get_member(1)
        self.assertTrue(m.called)
        self.assertIsInstance(member, Member)

    def test_create_api_key(self, m):
        register_uris({"fabman": ["create_api_key"]}, m)

        api_key = self.fabman.create_api_key(member=1, label="Quark's Bar Key")
        self.assertIsInstance(api_key, ApiKey)

    def test_create_booking(self, m):
        register_uris({"fabman": ["create_booking"]}, m)

        booking = self.fabman.create_booking(
            resource=1,
            fromDateTime="2021-01-01T12:00:00",
            toDateTime="2021-01-01T13:00:00",
            member=1,
        )
        self.assertIsInstance(booking, Booking)

    def test_create_charge(self, m):
        register_uris({"fabman": ["create_charge"]}, m)

        charge = self.fabman.create_charge(member=1, price=12.34)
        self.assertIsInstance(charge, Charge)

    def test_create_invoice(self, m):
        register_uris({"fabman": ["create_invoice"]}, m)

        invoice = self.fabman.create_invoice(member=1, date="2023-06-30")
        self.assertIsInstance(invoice, Invoice)

    def test_create_key_assignment(self, m):
        register_uris({"fabman": ["create_key_assignment"]}, m)

        resp = self.fabman.create_key_assignment(member=1)
        self.assertTrue(m.called)
        self.assertIsInstance(resp, requests.Response)
        self.assertTrue(resp.status_code == 201)

    def test_create_member(self, m):
        register_uris({"fabman": ["create_member"]}, m)

        member = self.fabman.create_member(
            account=1,
            firstName="Julian",
            lastName="Bashear",
            email="bashear@ds9.starfleet",
        )
        self.assertIsInstance(member, Member)

    def test_create_package(self, m):
        register_uris({"fabman": ["create_package"]}, m)

        package = self.fabman.create_package(account=1, name="Test Package")
        self.assertIsInstance(package, Package)

    def test_create_payment(self, m):
        register_uris({"fabman": ["create_payment"]}, m)

        payment = self.fabman.create_payment(member=1, total=12.34)
        self.assertIsInstance(payment, Payment)

    def test_create_resource(self, m):
        register_uris({"fabman": ["create_resource"]}, m)

        resource = self.fabman.create_resource(space=1, name="USS Defiant", type=1)
        self.assertIsInstance(resource, Resource)

    def test_create_resource_log(self, m):
        register_uris({"fabman": ["create_resource_log"]}, m)

        resource_log = self.fabman.create_resource_log(resource=1, member=1)
        self.assertIsInstance(resource_log, ResourceLog)

    def test_create_resource_type(self, m):
        register_uris({"fabman": ["create_resource_type"]}, m)

        resource_type = self.fabman.create_resource_type(account=1, name="Starship")
        self.assertIsInstance(resource_type, ResourceType)

    def test_create_space(self, m):
        register_uris({"fabman": ["create_space"]}, m)

        space = self.fabman.create_space(account=1, name="Operations Center")
        self.assertIsInstance(space, Space)

    def test_create_training_course(self, m):
        register_uris({"fabman": ["create_training_course"]}, m)

        course = self.fabman.create_training_course(account=1, name="Keiko's School")
        self.assertIsInstance(course, TrainingCourse)

    def test_create_webhook(self, m):
        register_uris({"fabman": ["create_webhook"]}, m)

        webhook = self.fabman.create_webhook(
            account=1,
            url="https://example.com/julians_door_cam",
            label="Garak's Test Webhook",
        )
        self.assertIsInstance(webhook, Webhook)

    def test_get_account(self, m):
        register_uris({"fabman": ["get_account_by_id"]}, m)

        account = self.fabman.get_account(1)
        self.assertIsInstance(account, Account)
        self.assertTrue(hasattr(account, "id"))
        self.assertTrue(account.id == 1)

    def test_get_accounts(self, m):
        register_uris({"fabman": ["get_accounts"]}, m)

        accounts = self.fabman.get_accounts()
        self.assertIsInstance(accounts, PaginatedList)
        self.assertIsInstance(accounts[0], Account)
        self.assertTrue(hasattr(accounts[0], "id"))
        self.assertTrue(accounts[0].id == 1)

    def test_get_api_key(self, m):
        register_uris({"fabman": ["get_api_key_by_id"]}, m)

        api_key = self.fabman.get_api_key(1)
        self.assertIsInstance(api_key, ApiKey)
        self.assertTrue(hasattr(api_key, "id"))
        self.assertTrue(api_key.id == 1)

    def test_get_api_keys(self, m):
        register_uris({"fabman": ["get_api_keys"]}, m)

        api_keys = self.fabman.get_api_keys()
        self.assertIsInstance(api_keys, PaginatedList)
        self.assertIsInstance(api_keys[0], ApiKey)
        self.assertTrue(hasattr(api_keys[0], "id"))
        self.assertTrue(api_keys[0].id == 1)

    def test_get_booking(self, m):
        register_uris({"fabman": ["get_booking_by_id"]}, m)

        bookings = self.fabman.get_booking(1)
        self.assertIsInstance(bookings, Booking)
        self.assertTrue(hasattr(bookings, "id"))
        self.assertTrue(bookings.id == 1)

    def test_get_bookings(self, m):
        register_uris({"fabman": ["get_bookings"]}, m)

        bookings = self.fabman.get_bookings()
        self.assertIsInstance(bookings, PaginatedList)
        self.assertIsInstance(bookings[0], Booking)
        self.assertTrue(hasattr(bookings[0], "id"))
        self.assertTrue(bookings[0].id == 1)

    def test_get_charge(self, m):
        register_uris({"fabman": ["get_charge_by_id"]}, m)

        charge = self.fabman.get_charge(1)
        self.assertIsInstance(charge, Charge)
        self.assertTrue(hasattr(charge, "id"))
        self.assertTrue(charge.id == 1)

    def test_get_charges(self, m):
        register_uris({"fabman": ["get_charges"]}, m)

        charges = self.fabman.get_charges()
        self.assertIsInstance(charges, PaginatedList)
        self.assertIsInstance(charges[0], Charge)
        self.assertTrue(hasattr(charges[0], "id"))
        self.assertTrue(charges[0].id == 1)

    def test_get_invoice(self, m):
        register_uris({"fabman": ["get_invoice_by_id"]}, m)

        invoice = self.fabman.get_invoice(1)
        self.assertIsInstance(invoice, Invoice)
        self.assertTrue(hasattr(invoice, "id"))
        self.assertTrue(invoice.id == 1)

    def test_invoices(self, m):
        register_uris({"fabman": ["get_invoices"]}, m)

        invoices = self.fabman.get_invoices()
        self.assertIsInstance(invoices, PaginatedList)
        self.assertIsInstance(invoices[0], Invoice)
        self.assertTrue(hasattr(invoices[0], "id"))
        self.assertTrue(invoices[0].id == 1)

    def test_get_job(self, m):
        register_uris({"fabman": ["get_job_by_id"]}, m)

        job = self.fabman.get_job(1)
        self.assertIsInstance(job, Job)
        self.assertTrue(hasattr(job, "id"))
        self.assertTrue(job.id == 1)
        self.assertTrue(str(job) == "Job #1")

    def test_get_jobs(self, m):
        register_uris({"fabman": ["get_jobs"]}, m)

        jobs = self.fabman.get_jobs()
        self.assertIsInstance(jobs, PaginatedList)
        self.assertIsInstance(jobs[0], Job)
        self.assertTrue(hasattr(jobs[0], "id"))
        self.assertTrue(jobs[0].id == 1)

    def test_get_member(self, m):
        register_uris({"fabman": ["get_member_by_id"]}, m)

        member = self.fabman.get_member(1)
        self.assertIsInstance(member, Member)
        self.assertTrue(hasattr(member, "id"))
        self.assertTrue(hasattr(member, "firstName"))
        self.assertTrue(member.id == 1)

    def test_get_members(self, m):
        register_uris({"fabman": ["get_members"]}, m)

        members = self.fabman.get_members()
        self.assertIsInstance(members, PaginatedList)
        self.assertIsInstance(members[0], Member)
        self.assertTrue(hasattr(members[0], "id"))
        self.assertTrue(members[0].id == 1)

    def test_get_member_with_embed_str(self, m):
        """
        Tests the ability to embed a single resource as a string. Primarily tests
        the functionality of the Requester child object. Embedding does not need
        to be checked for other resources, as they all use the same Requester.
        """
        register_uris({"fabman": ["get_member_by_id_with_embed_str"]}, m)

        member = self.fabman.get_member(1, embed="memberPackages")
        self.assertIsInstance(member, Member)
        self.assertTrue(hasattr(member, "id"))
        self.assertTrue(hasattr(member, "_embedded"))

    def test_get_member_with_embed_single_list(self, m):
        """
        Tests the ability to embed a single resource in a list. Primarily tests
        the functionality of the Requester child object. Embedding does not need
        to be checked for other resources, as they all use the same Requester.
        """
        register_uris({"fabman": ["get_member_by_id_with_embed_str"]}, m)

        member = self.fabman.get_member(1, embed=["memberPackages"])
        self.assertIsInstance(member, Member)
        self.assertTrue(hasattr(member, "id"))
        self.assertTrue(hasattr(member, "_embedded"))

        self.assertTrue(
            "memberPackages" in member._embedded
        )  # pylint: disable=protected-access

    def test_get_member_with_embed_multi_list(self, m):
        """
        Tests the ability to embed multiple resource in a list. Primarily tests
        the functionality of the Requester child object. Embedding does not need
        to be checked for other resources, as they all use the same Requester.
        """
        register_uris({"fabman": ["get_member_by_id_with_embed_list_multi"]}, m)

        member = self.fabman.get_member(1, embed=["memberPackages", "trainings"])
        self.assertIsInstance(member, Member)
        self.assertTrue(hasattr(member, "id"))
        self.assertTrue(hasattr(member, "_embedded"))

        embeds = member._embedded  # pylint: disable=protected-access
        self.assertTrue("memberPackages" in embeds)
        self.assertTrue("trainings" in embeds)

    def test_get_package(self, m):
        register_uris({"fabman": ["get_package_by_id"]}, m)

        package = self.fabman.get_package(1)
        self.assertIsInstance(package, Package)
        self.assertTrue(hasattr(package, "id"))
        self.assertTrue(package.id == 1)

    def test_get_packages(self, m):
        register_uris({"fabman": ["get_packages"]}, m)

        packages = self.fabman.get_packages()
        self.assertIsInstance(packages, PaginatedList)
        self.assertIsInstance(packages[0], Package)
        self.assertTrue(hasattr(packages[0], "id"))
        self.assertTrue(packages[0].id == 1)

    def test_get_payment(self, m):
        register_uris({"fabman": ["get_payment_by_id"]}, m)

        payment = self.fabman.get_payment(1)
        self.assertIsInstance(payment, Payment)
        self.assertTrue(hasattr(payment, "id"))
        self.assertTrue(payment.id == 1)

    def test_get_payments(self, m):
        register_uris({"fabman": ["get_payments"]}, m)

        payments = self.fabman.get_payments()
        self.assertIsInstance(payments, PaginatedList)
        self.assertIsInstance(payments[0], Payment)
        self.assertTrue(hasattr(payments[0], "id"))
        self.assertTrue(payments[0].id == 1)

    def test_get_resource(self, m):
        register_uris({"fabman": ["get_resource_by_id"]}, m)

        resource = self.fabman.get_resource(1)
        self.assertIsInstance(resource, Resource)
        self.assertTrue(hasattr(resource, "id"))
        self.assertTrue(hasattr(resource, "name"))

    def test_get_resources(self, m):
        register_uris({"fabman": ["get_resources"]}, m)

        resources = self.fabman.get_resources()
        self.assertIsInstance(resources, PaginatedList)
        self.assertIsInstance(resources[0], Resource)
        self.assertTrue(hasattr(resources[0], "id"))
        self.assertTrue(resources[0].id == 1)

    def test_get_resource_log(self, m):
        register_uris({"fabman": ["get_resource_log_by_id"]}, m)

        resources = self.fabman.get_resource_log(1)
        self.assertIsInstance(resources, ResourceLog)
        self.assertTrue(hasattr(resources, "id"))
        self.assertTrue(resources.id == 1)

    def test_get_resource_logs(self, m):
        register_uris({"fabman": ["get_resource_logs"]}, m)

        resources = self.fabman.get_resource_logs()
        self.assertIsInstance(resources, PaginatedList)
        self.assertIsInstance(resources[0], ResourceLog)
        self.assertTrue(hasattr(resources[0], "id"))
        self.assertTrue(resources[0].id == 1)

    def test_get_resource_types(self, m):
        register_uris({"fabman": ["get_resource_types"]}, m)

        resource_types = self.fabman.get_resource_types()
        self.assertIsInstance(resource_types, PaginatedList)
        self.assertTrue(hasattr(resource_types[0], "id"))
        self.assertTrue(resource_types[0].id == 1)

    def test_get_space(self, m):
        register_uris({"fabman": ["get_space_by_id"]}, m)

        space = self.fabman.get_space(1)
        self.assertIsInstance(space, Space)
        self.assertTrue(hasattr(space, "id"))
        self.assertTrue(space.id == 1)

    def test_get_spaces(self, m):
        register_uris({"fabman": ["get_spaces"]}, m)

        spaces = self.fabman.get_spaces()
        self.assertIsInstance(spaces, PaginatedList)
        self.assertIsInstance(spaces[0], Space)
        self.assertTrue(hasattr(spaces[0], "id"))
        self.assertTrue(spaces[0].id == 1)

    def test_get_training_course(self, m):
        register_uris({"fabman": ["get_training_course_by_id"]}, m)

        training_course = self.fabman.get_training_course(1)
        self.assertIsInstance(training_course, TrainingCourse)
        self.assertTrue(hasattr(training_course, "id"))
        self.assertTrue(training_course.id == 1)

    def test_get_training_courses(self, m):
        register_uris({"fabman": ["get_training_courses"]}, m)

        training_courses = self.fabman.get_training_courses()
        self.assertIsInstance(training_courses, PaginatedList)
        self.assertIsInstance(training_courses[0], TrainingCourse)
        self.assertTrue(hasattr(training_courses[0], "id"))
        self.assertTrue(training_courses[0].id == 1)

    def test_get_user_me(self, m):
        register_uris({"fabman": ["get_user_me"]}, m)

        member = self.fabman.get_user()
        self.assertIsInstance(member, Member)
        self.assertTrue(hasattr(member, "id"))
        self.assertTrue(hasattr(member, "firstName"))
        self.assertTrue(hasattr(member, "account"))

    def test_get_webhook(self, m):
        register_uris({"fabman": ["get_webhook_by_id"]}, m)

        webhook = self.fabman.get_webhook(1)
        self.assertIsInstance(webhook, Webhook)
        self.assertTrue(hasattr(webhook, "id"))
        self.assertTrue(webhook.id == 1)

    def test_get_webhooks(self, m):
        register_uris({"fabman": ["get_webhooks"]}, m)

        webhooks = self.fabman.get_webhooks()
        self.assertIsInstance(webhooks, PaginatedList)
        self.assertIsInstance(webhooks[0], Webhook)
        self.assertTrue(hasattr(webhooks[0], "id"))
        self.assertTrue(webhooks[0].id == 1)


class TestFabmanObject(unittest.TestCase):
    def test_repr(self):
        requester = Requester(settings.BASE_URL_WITH_VERSION, settings.API_KEY)
        obj = FabmanObject(requester, {"id": 1})
        self.assertEqual(repr(obj), "<FabmanObject id=1>")


if __name__ == "__main__":
    unittest.main()



"""Defines the Payment object"""

import requests

from fabman.fabman_object import FabmanObject


class PaymentRequest(FabmanObject):
    """Simple class to hold payment request information"""

    def __str__(self):
        return f"PaymentRequest for Payment #{self.payment_id}"


class Payment(FabmanObject):
    """Defines the Payment object for interacting with payments on the Fabman API"""

    def __str__(self) -> str:
        return f"Payment #{self.id}: {self.notes}"

    def delete(self, **kwargs) -> requests.Response:
        """
        Deletes a payment. *WARNING: This is irreversible.*

        :calls: "DELETE /payments/{id}" \
		<https://fabman.io/api/v1/documentation#/payments/deletePaymentsId>

        :returns: response information of call
        :rtype: requests.Response
        """
        uri = f"/payments/{self.id}"

        response = self._requester.request("DELETE", uri, _kwargs=kwargs)

        return response

    def request_payment(self, **kwargs) -> PaymentRequest:
        """
        Requests a payment from the customer.

        :calls: "POST /payments/{id}/request-payment" \
		<https://fabman.io/api/v1/documentation#/payments/postPaymentsIdRequestpayment>
        
        :return: response information of the call
        :rtype: PaymentRequest
        """
        uri = f"/payments/{self.id}/request-payment"

        response = self._requester.request("POST", uri, _kwargs=kwargs)

        data = response.json()
        data.update({"payment_id": self.id})

        return PaymentRequest(self._requester, data)

    def update(self, **kwargs) -> None:
        """
        Updates a payment. Attributes are modified in place using updated information
        from the API.

        :calls: "PUT /payments/{id}" \
		<https://fabman.io/api/v1/documentation#/payments/putPaymentsId>
  
        :return: None -- attributes are updated in place
        :rtype: None
        """

        uri = f"/payments/{self.id}"

        kwargs.update({"lockVersion": self.lockVersion})

        response = self._requester.request("PUT", uri, _kwargs=kwargs)

        data = response.json()

        for attr, val in data.items():
            setattr(self, attr, val)



"""Base Fabman Object for all other returned objects"""

from typing import Any

import fabman.requester


class FabmanObject(object):
    """
    Base class for all classes representing objects returned by the API

    This makes a call to `fabman.fabman_object.FabmanObject.set_attributes`
    to dynamically construct this object's attributes with a JSON object

    Based off of canvasapi.canvas_object.CanvasObject found at
    https://github.com/ucfopen/canvasapi/blob/develop/canvasapi/canvas_object.py
    """

    def __getattribute__(self, __name: str) -> Any:
        return super(FabmanObject, self).__getattribute__(__name)

    def __init__(self, requester: fabman.requester.Requester, attributes: dict) -> None:
        """Initialize the Object. Stores the requester method to interact
        with the API for further calls

        :param requester: The :code:`Requester` object to make requests with
        :type requester: :code:`Requester`
        :param attributes: The attributes to initialize this object with
        :type attributes: dict
        """

        self._requester = requester
        self.set_attributes(attributes)
        if "_embedded" not in self.__dict__:
            self._embedded = {}

    def __repr__(self) -> str:
        classname = self.__class__.__name__
        attrs = ", ".join(
            [
                f"{k}={v}"
                for k, v in self.__dict__.items()
                if not k.startswith("_") and k != "attributes"
            ]
        )

        return f"<{classname} {attrs}>"

    def set_attributes(self, attributes: dict):
        """
        Loads the Object with the specified attributes. Typically, these attributes
        will be returned by the API when the object is created. Can also be used when
        update 'PUT' requests are made to the API.

        :param attributes: The attributes to initialize this object with
        :type attributes: dict
        """
        for attr, val in attributes.items():
            setattr(self, attr, val)



"""Tests for the TrainingCourse class."""
# pylint: disable=missing-docstring, invalid-name, unused-argument

import unittest

import requests_mock

from fabman import Fabman
from fabman.training_course import TrainingCourse
from tests import settings
from tests.util import register_uris, validate_update


@requests_mock.Mocker()
class TestTrainingCourse(unittest.TestCase):
    def setUp(self):
        self.fabman = Fabman(settings.API_KEY)

        with requests_mock.Mocker() as m:
            register_uris({"fabman": ["get_training_course_by_id"]}, m)

            self.training_course: TrainingCourse = self.fabman.get_training_course(1)

    def test_sanity(self, m):
        self.assertEqual(1, 1)

    def test_instance(self, m):
        self.assertIsInstance(self.training_course, TrainingCourse)

    def test_str(self, m):
        string = str(self.training_course)
        self.assertTrue(string == "TrainingCourse #1: Keiko's Botany Class")

    def test_delete(self, m):
        register_uris({"training_course": ["delete"]}, m)

        response = self.training_course.delete()

        self.assertTrue(response)

    def test_update(self, m):
        m.register_uri(
            "PUT",
            f"{settings.BASE_URL_WITH_VERSION}/training-courses/1",
            text=validate_update,
            status_code=200,
        )

        self.training_course.update(title="Bajoran Worker Training")
        self.assertTrue(m.called)



"""Tests for the Requester Class"""
# pylint: disable=missing-docstring, invalid-name, unused-argument, protected-access
import unittest

import requests_mock

from fabman.exceptions import (
    BadRequest,
    Conflict,
    FabmanException,
    ForbiddenError,
    InvalidAccessToken,
    RateLimitExceeded,
    Unauthorized,
    UnprocessableEntity,
)
from fabman.requester import Requester
from tests import settings
from tests.util import test_exceptions


@requests_mock.Mocker()
class TestRequester(unittest.TestCase):
    def setUp(self):
        self.requester = Requester(settings.BASE_URL_WITH_VERSION, settings.API_KEY)

    def test_sanity(self, m):
        self.assertEqual(1, 1)

    def test_instance(self, m):
        self.assertIsInstance(self.requester, Requester)

    def test_get(self, m):
        m.register_uri(
            "GET",
            f"{settings.BASE_URL_WITH_VERSION}/test",
            status_code=200,
        )

        resp = self.requester.request("GET", "/test")

        self.assertTrue(resp.status_code == 200)

    def test_post(self, m):
        m.register_uri(
            "POST",
            f"{settings.BASE_URL_WITH_VERSION}/test",
            status_code=201,
        )

        resp = self.requester.request("POST", "/test")

        self.assertTrue(resp.status_code == 201)

    def test_put(self, m):
        m.register_uri("PUT", f"{settings.BASE_URL_WITH_VERSION}/test", status_code=200)

        resp = self.requester.request("PUT", "/test")

        self.assertTrue(resp.status_code == 200)

    def test_delete(self, m):
        m.register_uri(
            "DELETE", f"{settings.BASE_URL_WITH_VERSION}/test", status_code=204
        )

        resp = self.requester.request("DELETE", "/test")

        self.assertTrue(resp.status_code == 204)

    def test_invalid(self, m):
        m.register_uri(
            "GAHBAGE", f"{settings.BASE_URL_WITH_VERSION}/test", status_code=420
        )

        with self.assertRaises(ValueError, msg="Invalid method GAHBAGE"):
            self.requester.request("GAHBAGE", "/test")

    def test_cache(self, m):
        m.register_uri("GET", f"{settings.BASE_URL_WITH_VERSION}/test", status_code=200)
        for i in range(4):
            self.assertTrue(len(self.requester._Requester__cache) == i)
            self.requester.request("GET", "/test")
            self.assertTrue(len(self.requester._Requester__cache) == i + 1)

        self.requester.request("GET", "/test")
        self.assertTrue(len(self.requester._Requester__cache) == 4)

    def test_400(self, m):
        m.register_uri(
            "GET",
            f"{settings.BASE_URL_WITH_VERSION}/test_400",
            text=test_exceptions,
            status_code=400,
        )

        with self.assertRaises(BadRequest, msg="Bad Request"):
            self.requester.request("GET", "/test_400")

            self.assertTrue(m.called)

    def test_401(self, m):
        m.register_uri(
            "GET",
            f"{settings.BASE_URL_WITH_VERSION}/test_401_invalid",
            text=test_exceptions,
            status_code=401,
        )

        with self.assertRaises(InvalidAccessToken, msg="Invalid access token"):
            self.requester.request("GET", "/test_401_invalid")

            self.assertTrue(m.called)

    def test_401_unauthorized(self, m):
        m.register_uri(
            "GET",
            f"{settings.BASE_URL_WITH_VERSION}/test_401_unauthorized",
            text=test_exceptions,
            status_code=401,
        )

        with self.assertRaises(Unauthorized, msg="Unauthorized"):
            self.requester.request("GET", "/test_401_unauthorized")

            self.assertTrue(m.called)

    def test_403(self, m):
        m.register_uri(
            "GET",
            f"{settings.BASE_URL_WITH_VERSION}/test_403",
            text=test_exceptions,
            status_code=403,
        )

        with self.assertRaises(ForbiddenError, msg="Forbidden"):
            self.requester.request("GET", "/test_403")

            self.assertTrue(m.called)

    def test_404(self, m):
        m.register_uri(
            "GET",
            f"{settings.BASE_URL_WITH_VERSION}/test_404",
            text=test_exceptions,
            status_code=404,
        )

        with self.assertRaises(FabmanException, msg="Not Found"):
            self.requester.request("GET", "/test_404")

            self.assertTrue(m.called)

    def test_409(self, m):
        m.register_uri(
            "GET",
            f"{settings.BASE_URL_WITH_VERSION}/test_409",
            text=test_exceptions,
            status_code=409,
        )

        with self.assertRaises(Conflict, msg="Conflict"):
            self.requester.request("GET", "/test_409")

            self.assertTrue(m.called)

    def test_422(self, m):
        m.register_uri(
            "GET",
            f"{settings.BASE_URL_WITH_VERSION}/test_422",
            text=test_exceptions,
            status_code=422,
        )

        with self.assertRaises(UnprocessableEntity, msg="Unprocessable Entity"):
            self.requester.request("GET", "/test_422")

            self.assertTrue(m.called)

    def test_429(self, m):
        m.register_uri(
            "GET",
            f"{settings.BASE_URL_WITH_VERSION}/test_429",
            text=test_exceptions,
            status_code=429,
        )

        with self.assertRaises(RateLimitExceeded, msg="Rate Limit Exceeded"):
            self.requester.request("GET", "/test_429")

            self.assertTrue(m.called)

    def test_4xx(self, m):
        m.register_uri(
            "GET",
            f"{settings.BASE_URL_WITH_VERSION}/test_420",
            text=test_exceptions,
            status_code=420,
        )

        with self.assertRaises(FabmanException, msg="Unknown 420"):
            self.requester.request("GET", "/test_420")

            self.assertTrue(m.called)



"""Tests for the Account class."""
# pylint: disable=missing-docstring, invalid-name, unused-argument

import unittest

import requests_mock

from fabman import Fabman
from fabman.account import Account, PaymentInfo
from tests import settings
from tests.util import register_uris, validate_update


@requests_mock.Mocker()
class TestAccount(unittest.TestCase):
    def setUp(self) -> None:
        self.fabman = Fabman(settings.API_KEY)

        with requests_mock.Mocker() as m:
            register_uris({"fabman": ["get_account_by_id"]}, m)

            self.account: Account = self.fabman.get_account(1)

    def test_instance(self, m):
        self.assertIsInstance(self.account, Account)

    def test_str(self, m):
        string = str(self.account)
        self.assertTrue(string == "Account #1: Tarok Nor")

    def test_get_payment_info(self, m):
        register_uris({"accounts": ["get_payment_info"]}, m)

        info = self.account.get_payment_info()
        self.assertIsInstance(info, PaymentInfo)
        self.assertEqual(info.id, 1)
        self.assertEqual(info.account_id, 1)

        self.assertEqual(str(info), "PaymentInfo for Account #1: stripe")

    def test_update(self, m):
        m.register_uri(
            "PUT",
            f"{settings.BASE_URL_WITH_VERSION}/accounts/1",
            text=validate_update,
            status_code=200,
        )

        self.account.update(name="New Name")
        self.assertTrue(m.called)



"""Tests for the Package class."""
# pylint: disable=missing-docstring, invalid-name, unused-argument

import unittest

import requests
import requests_mock

from fabman import Fabman
from fabman.package import Package, PackageCredit, PackagePermission
from fabman.paginated_list import PaginatedList
from tests import settings
from tests.util import register_uris, validate_update


@requests_mock.Mocker()
class TestPackage(unittest.TestCase):
    def setUp(self):
        self.fabman = Fabman(settings.API_KEY)

        with requests_mock.Mocker() as m:
            register_uris({"fabman": ["get_package_by_id"]}, m)

            self.package: Package = self.fabman.get_package(1)

    def test_instance(self, m):
        self.assertIsInstance(self.package, Package)

    def test_str(self, m):
        string = str(self.package)

        self.assertTrue(string == "Package #1: Test Package")

    def test_create_credit(self, m):
        register_uris({"package": ["create_credit"]}, m)

        credit = self.package.create_credit()
        self.assertIsInstance(credit, PackageCredit)
        self.assertTrue(hasattr(credit, "package_id"))

    def test_create_permission(self, m):
        register_uris({"package": ["create_permission"]}, m)

        permission = self.package.create_permission()
        self.assertIsInstance(permission, PackagePermission)
        self.assertTrue(hasattr(permission, "package_id"))

    def test_delete(self, m):
        register_uris({"package": ["delete"]}, m)

        resp = self.package.delete()
        self.assertIsInstance(resp, requests.Response)
        self.assertTrue(resp.status_code == 204)

    def test_get_credit(self, m):
        register_uris({"package": ["get_credit"]}, m)

        credit = self.package.get_credit(1)

        self.assertIsInstance(credit, PackageCredit)
        self.assertTrue(hasattr(credit, "id"))
        self.assertTrue(credit.id == 1)

    def test_get_credits(self, m):
        register_uris({"package": ["get_credits"]}, m)

        _credits = self.package.get_credits()
        self.assertIsInstance(_credits, PaginatedList)
        self.assertIsInstance(_credits[0], PackageCredit)
        self.assertTrue(hasattr(_credits[0], "id"))
        self.assertTrue(_credits[0].id == 1)

    def test_get_permission(self, m):
        register_uris({"package": ["get_permission"]}, m)

        permission = self.package.get_permission(1)
        self.assertIsInstance(permission, PackagePermission)
        self.assertTrue(hasattr(permission, "id"))
        self.assertTrue(permission.id == 1)

    def test_get_permissions(self, m):
        register_uris({"package": ["get_permissions"]}, m)

        permissions = self.package.get_permissions()
        self.assertIsInstance(permissions, PaginatedList)
        self.assertIsInstance(permissions[0], PackagePermission)
        self.assertTrue(hasattr(permissions[0], "id"))
        self.assertTrue(permissions[0].id == 1)

    def test_update(self, m):
        m.register_uri(
            "PUT",
            f"{settings.BASE_URL_WITH_VERSION}/packages/1",
            text=validate_update,
            status_code=200,
        )

        self.package.update(name="Test Package")
        self.assertTrue(m.called)


@requests_mock.Mocker()
class TestPackageCredit(unittest.TestCase):
    def setUp(self) -> None:
        self.fabman = Fabman(settings.API_KEY)

        with requests_mock.Mocker() as m:
            register_uris(
                {"fabman": ["get_package_by_id"], "package": ["get_credit"]}, m
            )

            self.package = self.fabman.get_package(1)
            self.credit = self.package.get_credit(1)

    def test_instance(self, m):
        self.assertIsInstance(self.credit, PackageCredit)

    def test_str(self, m):
        string = str(self.credit)
        self.assertTrue(string == "PackageCredit #1, Package #1: booking")

    def test_delete(self, m):
        register_uris({"package": ["delete_credit"]}, m)

        resp = self.credit.delete()
        self.assertTrue(m.called)
        self.assertIsInstance(resp, requests.Response)
        self.assertTrue(resp.status_code == 204)

    def test_update(self, m):
        m.register_uri(
            "PUT",
            f"{settings.BASE_URL_WITH_VERSION}/packages/1/credits/1",
            text=validate_update,
            status_code=200,
        )

        self.credit.update(scope="spoofing")
        self.assertTrue(m.called)


@requests_mock.Mocker()
class TestPackagePermission(unittest.TestCase):
    def setUp(self):
        self.fabman = Fabman(settings.API_KEY)

        with requests_mock.Mocker() as m:
            register_uris(
                {"fabman": ["get_package_by_id"], "package": ["get_permission"]}, m
            )

            self.package = self.fabman.get_package(1)
            self.permission = self.package.get_permission(1)

    def test_instance(self, m):
        self.assertIsInstance(self.permission, PackagePermission)

    def test_str(self, m):
        string = str(self.permission)
        self.assertTrue(string == "PackagePermission #1, Package #1: everything")

    def test_delete(self, m):
        register_uris({"package": ["delete_permission"]}, m)

        resp = self.permission.delete()
        self.assertTrue(m.called)
        self.assertIsInstance(resp, requests.Response)
        self.assertTrue(resp.status_code == 204)

    def test_update(self, m):
        m.register_uri(
            "PUT",
            f"{settings.BASE_URL_WITH_VERSION}/packages/1/permissions/1",
            text=validate_update,
            status_code=200,
        )

        self.permission.update(scope="spoofing")
        self.assertTrue(m.called)



"""Tests for the EmbeddedList class"""
# pylint: disable=missing-docstring, invalid-name, unused-argument

import unittest

import requests_mock

from fabman.embedded_list import EmbeddedList
from fabman.member import Member
from fabman.paginated_list import PaginatedList
from fabman.requester import Requester
from tests import settings


@requests_mock.Mocker()
class TestEmbeddedList(unittest.TestCase):
    def setUp(self):
        members = [
            {"id": 1, "firstName": "Julian", "lastName": "Bashear"},
            {"id": 2, "firstName": "Kira", "lastName": "Nerys"},
            {"id": 3, "firstName": "Benjamin", "lastName": "Sisko"},
            {"id": 4, "firstName": "Jadzia", "lastName": "Dax"},
            {"id": 5, "firstName": "Miles", "lastName": "O'Brien"},
            {"id": 6, "firstName": "Odo", "lastName": "Odo"},
            {"id": 7, "firstName": "Quark", "lastName": "Quark"},
            {"id": 8, "firstName": "Worf", "lastName": "Worf"},
            {"id": 9, "firstName": "Jake", "lastName": "Sisko"},
            {"id": 10, "firstName": "Nog", "lastName": "Nog"},
            {"id": 11, "firstName": "Rom", "lastName": "Rom"},
            {"id": 12, "firstName": "Leeta", "lastName": "Leeta"},
            {"id": 13, "firstName": "Elim", "lastName": "Garak"},
            {"id": 14, "firstName": "Martok", "lastName": "Martok"},
            {"id": 15, "firstName": "Kasidy", "lastName": "Yates"},
            {"id": 16, "firstName": "Ezri", "lastName": "Dax"},
            {"id": 17, "firstName": "Weyoun", "lastName": "Weyoun"},
            {"id": 18, "firstName": "Damar", "lastName": "Damar"},
            {"id": 19, "firstName": "Luaran", "lastName": "Luaran"},
        ]

        self._requester = Requester("https://fabman.io/api/v1", settings.API_KEY)
        extra_attribs = {"spaceId": 1, "account": 1}

        self.embedded_list = EmbeddedList(
            Member,
            members,
            self._requester,
            "GET",
            "/members",
            extra_attribs=extra_attribs,
        )

    def test_instance(self, m):
        self.assertIsInstance(self.embedded_list, EmbeddedList)

    def test_repr(self, m):
        self.assertEqual(
            repr(self.embedded_list), "<EmbeddedList of type Member with 19 elements>"
        )

    def test_iteration(self, m):
        for i, element in enumerate(self.embedded_list):
            self.assertIsInstance(element, Member)
            self.assertEqual(element.id, i + 1)

    def test_refresh(self, m):
        members = self.embedded_list.get_live_data()
        self.assertIsInstance(members, PaginatedList)

    def test_refresh_no_method(self, m):
        self.embedded_list._request_method = None
        with self.assertRaises(ValueError):
            self.embedded_list.get_live_data()
        self.embedded_list._request_method = ""
        with self.assertRaises(ValueError):
            self.embedded_list.get_live_data()

    def test_refresh_no_endpoint(self, m):
        self.embedded_list._refresh_endpoint = None
        with self.assertRaises(ValueError):
            self.embedded_list.get_live_data()
        self.embedded_list._refresh_endpoint = ""
        with self.assertRaises(ValueError):
            self.embedded_list.get_live_data()



"""Define the Invoice Object"""

import requests

from fabman.fabman_object import FabmanObject


class InvoiceDetails(FabmanObject):
    """Simple Class to handle Invoice Details"""

    def __str__(self):
        return f"Invoice #{self.id}: {self.totalExcludingTax} # Charges: {len(self.charges)}"


class Invoice(FabmanObject):
    """Simple Class to handle Invoices"""

    def __str__(self):
        return f"Invoice #{self.id}: {self.total} {self.state}"

    def cancel(self) -> requests.Response:
        """
        Cancels an invoice. Takes no arguments.

        :calls: "POST /invoices/{id}/cancel" \
		<https://fabman.io/api/v1/documentation#/invoices/postInvoicesIdCancel>
  
        :returns: An empty dict if successful.
        :type: dict
        """

        data = {"lockVersion": self.lockVersion}

        uri = f"/invoices/{self.id}/cancel"
        response = self._requester.request("POST", uri, _kwargs=data)

        return response.json()

    def details(self, **kwargs) -> InvoiceDetails:
        """
        Returns details about the invoice. Takes no arguments.

        :calls: "GET /invoices/{id}/details" \
		<https://fabman.io/api/v1/documentation>
  
        :returns: A dict containing details about the invoice.
        :rtype: dict
        """

        if "details" in self._embedded:
            self._embedded["details"].update({"invoice_id": self.id})
            return InvoiceDetails(self._requester, self._embedded["details"])

        uri = f"/invoices/{self.id}/details"

        response = self._requester.request("GET", uri, _kwargs=kwargs)
        data = response.json()
        self._embedded["details"] = response.json()

        data.update({"invoice_id": self.id})
        return InvoiceDetails(self._requester, data)

    def update(self, **kwargs) -> None:
        """
        Updates the invoice. Attributes are updated in place with new information
        provided by the API

        :calls: "PUT /invoices/{id}" \
		<https://fabman.io/api/v1/documentation#/invoices/putInvoicesId>
  
        :returns: None -- updates attributes of the current object
        :rtype: None
        """

        kwargs.update({"lockVersion": self.lockVersion})

        uri = f"/invoices/{self.id}"
        response = self._requester.request("PUT", uri, _kwargs=kwargs)

        data = response.json()
        for attr, val in data.items():
            setattr(self, attr, val)



"""Defines and handles the Member object returned by the API"""
# pylint: disable=too-many-public-methods, line-too-long
from typing import List, Optional, Union

import requests

from fabman.embedded_list import EmbeddedList
from fabman.fabman_object import FabmanObject
from fabman.package import Package
from fabman.paginated_list import PaginatedList


class MemberBalanceItems(FabmanObject):
    """Simple Class to handle Member Balance Items"""


class MemberChange(FabmanObject):
    """Simple class to hold Member Changes"""

    def __str__(self):
        return f"Change #{self.id} for member #{self.member_id}"

    def delete(self, **kwargs) -> requests.Response:
        """
        Deletes a member change given change ID
        
        :calls: "DELETE /members/{id}/changes/{changeId}" \
        <https://fabman.io/api/v1/documentation#/members/deleteMembersIdChangesChangeid>
        
        :return: Response information of delete call
        :rtype: requests.Response
        """

        return self._requester.request(
            "DELETE", f"/members/{self.member_id}/changes/{self.id}", _kwargs=kwargs
        )


class MemberCreditUse(FabmanObject):
    """
    Simple class to hold Credit Use
    """

    def __str__(self):
        # TODO: Test this method. No current way of applying credits with account available to developers
        return f"{self.id}: {self.amount} - {self.description}"


class MemberCredit(FabmanObject):
    """
    Simple Class to handle operations on MemberCredits
    """

    def __str__(self):
        return f"{self.id}: {self.scope} - {self.amount}"

    def delete(self, **kwargs) -> requests.Response:
        """
        Deletes a credit from a user account. **WARNING: THIS CANNOT BE UNDONE.**

        :calls: "DELETE /members/{member_id}/credits/{credit_id}" \
		<https://fabman.io/api/v1/documentation#/members/deleteMembersIdCreditsCreditid>

        :returns: An empty dict if successful.
        :rtype: dict
        """
        response = self._requester.request(
            "DELETE", f"/members/{self.member_id}/credits/{self.id}", _kwargs=kwargs
        )

        return response

    def get_uses(self, **kwargs) -> List[MemberCreditUse]:
        """
        Retrieves a list of uses of the credit.

        :calls: "GET /members/{member_id}/credits/{credit_id}/uses" \
		<https://fabman.io/api/v1/documentation#/members/getMembersIdCreditsCreditidUses>
  
        :return: A list of uses of the credit.
        :rtype: list
        """
        response = self._requester.request(
            "GET", f"/members/{self.member_id}/credits/{self.id}/uses", _kwargs=kwargs
        )

        data = response.json()
        return [MemberCreditUse(self._requester, use) for use in data]

    def update(self, **kwargs) -> None:
        """
        Updates an existing credit in place

        :calls: "PUT /members/{id}/credits/{creditId}" \
		<https://fabman.io/api/v1/documentation#/members/putMembersIdCreditsCreditid>
  
        :returns: None
        :rtype: None
        """
        kwargs.update({"lockVersion": self.lockVersion})
        response = self._requester.request(
            "PUT", f"/members/{self.member_id}/credits/{self.id}", _kwargs=kwargs
        )

        data = response.json()

        for attr, val in data.items():
            setattr(self, attr, val)


class MemberDevice(FabmanObject):
    """Simple Class to represent a Member Device"""

    def __str__(self):
        return f"Device: {self.name}"


class MemberDeviceChange(FabmanObject):
    """Simple class to represent a Member Device Change"""

    def __str__(self) -> str:
        return f"DeviceChange #{self.id} for member #{self.member_id}"

    def delete(self, **kwargs) -> requests.Response:
        """
        Deletes a member device change given change ID
        
        :calls: "DELETE /members/{id}/device/changes/{changeId}" \
        <https://fabman.io/api/v1/documentation#/members/deleteMembersIdDeviceChangesChangeid>
        
        :returns: Response information of delete call
        :rtype: requests.Response
        """
        return self._requester.request(
            "DELETE",
            f"/members/{self.member_id}/device/changes/{self.id}",
            _kwargs=kwargs,
        )


class MemberInvitation(FabmanObject):
    """Simple class to represent a MemberInvitation"""


class MemberKey(FabmanObject):
    """
    Manage a member keycard object
    """

    def __str__(self):
        return f"{self.member} - {self.type}"

    def delete(self, **kwargs) -> requests.Response:
        """
        Deletes a member key. **WARNING: THIS CANNOT BE UNDONE.**

        :calls: "DELETE /members/{self.member}/key" \
		<https://fabman.io/api/v1/documentation#/members/deleteMembersIdKey>
  
        :returns: Response information of delete call
        :rtype: requests.Response
        """
        response = self._requester.request(
            "DELETE", f"/members/{self.member}/key", _kwargs=kwargs
        )

        return response

    def update(self, **kwargs) -> None:
        """
        Updates a member key and updates the MemberKey object in place with new
        data from the API.

        :calls: "PUT /member{self.member}/key" \
		<https://fabman.io/api/v1/documentation#/members/putMembersIdKey>
  
        :returns: None
        :rtype: None
        """

        kwargs.update({"lockVersion": self.lockVersion})

        response = self._requester.request(
            "PUT", f"/members/{self.member}/key", _kwargs=kwargs
        )

        data = response.json()

        for attr, val in data.items():
            setattr(self, attr, val)


class MemberPackage(FabmanObject):
    """
    Simple class for managing member packages
    """

    def __str__(self):
        return f"{self.id} - {self.package}"

    def delete(self, **kwargs) -> requests.Response:
        """Removes the package from the current user account. *WARNING: THIS CANNOT BE UNDONE.*

        :calls: "DELETE /members/{id}/packages/{memberPackageId}" \
		<https://fabman.io/api/v1/documentation#/members/deleteMembersIdPackagesMemberpackageid>

        :returns: Response information of delete call
        :rtype: requests.Response
        """

        response = self._requester.request(
            "DELETE", f"/members/{self.member_id}/packages/{self.id}", _kwargs=kwargs
        )

        return response

    def get_package(self, **kwargs) -> Package:
        """
        Gets information about the package

        :calls: "GET /packages/{id}" \
		<https://fabman.io/api/v1/documentation#/packages/getPackagesId>
        
        :returns: :code:`fabman.package.Package` object with Package details
        :rtype: fabman.package.Package
        """
        if "package" in self._embedded:
            data = self._embedded["package"]
        else:
            response = self._requester.request(
                "GET", f"/packages/{self.package}", _kwargs=kwargs
            )
            data = response.json()

        return Package(self._requester, data)

    def update(self, **kwargs) -> None:
        """
        Updates a member package and updates the MemberPackage object in place with new
        data from the API

        :calls: "PUT /members/{id}/packages/{memberPackageId}" \
		<https://fabman.io/api/v1/documentation#/members/putMembersIdPackagesMemberpackageid>
  
        :returns: None
        :rtype: None
        """
        kwargs.update({"lockVersion": self.lockVersion})

        response = self._requester.request(
            "PUT", f"/members/{self.member_id}/packages/{self.id}", _kwargs=kwargs
        )

        data = response.json()

        for attr, val in data.items():
            setattr(self, attr, val)


class MemberPaymentAccount(FabmanObject):
    """Simple object to hold Member Payment Accounts"""


class MemberPaymentMethod(FabmanObject):
    """Simple object to hold Member Payment Methods"""

    def __str__(self):
        return f"Member #{self.member_id} has a {self.type} payment method"

    def delete(self, **kwargs):
        """Deletes a member payment method
        
        :calls: "DELETE /members/{id}/payment-method" \
        <https://fabman.io/api/v1/documentation#/members/deleteMembersIdPaymentmethod>
        
        :return: Response information of the delete call
        :rtype: requests.Response
        """

        return self._requester.request(
            "DELETE", f"/members/{self.member_id}/payment-method", _kwargs=kwargs
        )


class MemberPaymentMethodMandatePreview(FabmanObject):
    """Simple object to hold Member Payment Method Mandate Previews"""


class MemberPrivileges(FabmanObject):
    """Simple object to hold Member Privileges"""

    def __str__(self):
        return f"Member #{self.member_id} is {self.privileges}"


class MemberTrainedResources(FabmanObject):
    """Simple object to hold Member Trained Resources"""

    def __str__(self):
        return f"Member #{self.member_id} is trained on {len(self.resources)} resources"


class MemberTraining(FabmanObject):
    """Simple object to handle Member Trainings"""

    def __str__(self):
        return f"MemberTraining #{self.id} for member #{self.member_id}"

    def delete(self, **kwargs):
        """Deletes a member training

        :calls: "DELETE /members/{id}/trainings/{trainingId}" \
        <https://fabman.io/api/v1/documentation#/members/deleteMembersIdTrainingsTrainingid>
        
        :return: Response Information of delete call
        :rtype: requests.Response
        """

        return self._requester.request(
            "DELETE", f"/members/{self.member_id}/trainings/{self.id}", _kwargs=kwargs
        )


class Member(FabmanObject):
    """
    Member object returned by the API. Provides access to all API calls that operate on a single member.
    """

    def __str__(self):
        return f"{self.id}: {self.firstName} {self.lastName}"

    def create_credit(self, **kwargs) -> MemberCredit:
        """
        Creates a member credit
        
        :calls: "POST /members/{id}/credits" \
		<https://fabman.io/api/v1/documentation#/members/postMembersIdCredits>

        :returns: :code:`fabman.member.MemberCredit` object with credit details
        :rtype: fabman.member.MemberCredit
        """

        kwargs.update({"lockVersion": self.lockVersion})

        response = self._requester.request(
            "POST",
            f"/members/{self.id}/credits",
            _kwargs=kwargs,
        )

        data = response.json()
        data.update({"member_id": self.id})

        return MemberCredit(self._requester, data)

    def create_key(self, **kwargs) -> MemberKey:
        """
        Creates a key for the member if one does not already exist. If member already has
        a key, use `update_key()`.

        :calls: "POST /members/{id}/key" \
		<https://fabman.io/api/v1/documentation#/members/postMembersIdKey>
  
        :returns: :code:`fabman.member.MemberKey` object with key details
        :rtype: fabman.member.MemberKey
        """

        response = self._requester.request(
            "POST", f"/members/{self.id}/key", _kwargs=kwargs
        )

        return MemberKey(self._requester, response.json())

    def create_training(self, **kwargs) -> MemberTraining:
        """Creates a Training object which links a member to a training course.

        :return: MemberTraining object
        :rtype: MemberTraining
        """

        response = self._requester.request(
            "POST", f"/members/{self.id}/trainings", _kwargs=kwargs
        )

        return MemberTraining(self._requester, response.json())

    def delete(self, **kwargs) -> requests.Response:
        """
        Deletes a member
        
        :calls: "DELETE /members/{id}" \
		<https://fabman.io/api/v1/documentation#/members/deleteMembersId>
  
        :returns: Response information of delete call
        :rtype: requests.Response
        """
        return self._requester.request(
            "DELETE",
            f"/members/{self.id}",
            _kwargs=kwargs,
        )

    def get_balance_items(self, **kwargs) -> MemberBalanceItems:
        """
        Retrieves the balance items of a member
        
        :calls: "GET /members/{id}/balance-items" \
		<https://fabman.io/api/v1/documentation#/members/getMembersIdBalanceitems>
  
        :returns: Response information of get call
        :rtype: MemberBalanceItems
        """
        response = self._requester.request(
            "GET",
            f"/members/{self.id}/balance-items",
            _kwargs=kwargs,
        )

        data = response.json()
        data.update({"member_id": self.id})

        return MemberBalanceItems(self._requester, data)

    def get_changes(self, **kwargs) -> List[MemberChange]:
        """
        Retrieves the changes of a member
        
        :calls: "GET /members/{id}/changes" \
		<https://fabman.io/api/v1/documentation#/members/getMembersIdChanges>
  
        :returns: List of changes of a member
        :rtype: list[MemberChange]
        """
        response = self._requester.request(
            "GET",
            f"/members/{self.id}/changes",
            _kwargs=kwargs,
        )
        data = response.json()
        for change in data:
            change.update({"member_id": self.id})

        return [MemberChange(self._requester, x) for x in data]

    def get_credits(self, **kwargs) -> PaginatedList:
        """
        Retrieves the credits of a member
        :calls: "GET /members/{id}/credits" \
		<https://fabman.io/api/v1/documentation#/members/getMembersIdCredits>
  
        :returns: List of credits of a member
        :rtype: fabman.paginated_list.PaginatedList
        """
        return PaginatedList(
            MemberCredit,
            self._requester,
            "GET",
            f"/members/{self.id}/credits",
            extra_attribs={"member_id": self.id},
            **kwargs,
        )

    def get_credit(self, credit_id: int, **kwargs) -> MemberCredit:
        """
        Retrieves a credit of a member
        
        :calls: "GET /members/{id}/credits/{creditId}" \
		<https://fabman.io/api/v1/documentation#/members/getMembersIdCreditsCreditid>
  
        :param credit_id: ID of the credit to retrieve
        :type credit_id: int
        :returns: :code:`fabman.member.MemberCredit` object with credit details
        :rtype: fabman.member.MemberCredit
        """
        response = self._requester.request(
            "GET",
            f"/members/{self.id}/credits/{credit_id}",
            _kwargs=kwargs,
        )

        data = response.json()
        data.update({"member_id": self.id})

        return MemberCredit(self._requester, data)

    def get_device(self, **kwargs) -> MemberDevice:
        """
        Retrieves the device of a member used to authenticate on a bridge.
        
        :calls: "GET /members/{id}/devices" \
		<https://fabman.io/api/v1/documentation#/members/getMembersIdDevice>
  
        :returns: Device information
        :rtype: dict
        """
        if "device" in self._embedded:
            data = self._embedded["device"]
        else:
            response = self._requester.request(
                "GET",
                f"/members/{self.id}/device",
                _kwargs=kwargs,
            )

            data = response.json()
        data.update({"member_id": self.id})

        return MemberDevice(self._requester, data)

    def get_device_changes(self, **kwargs) -> List[MemberDeviceChange]:
        """
        Retrieves the device changes of a member
        
        :calls: "GET /members/{id}/device/changes" \
		<https://fabman.io/api/v1/documentation#/members/getMembersIdDeviceChanges>

        :returns: List of device changes of a member
        :rtype: list[MemberDeviceChange]
        """
        response = self._requester.request(
            "GET",
            f"/members/{self.id}/device/changes",
            _kwargs=kwargs,
        )
        data = response.json()
        for el in data:
            el.update({"member_id": self.id})

        return [MemberDeviceChange(self._requester, change) for change in data]

    def get_device_change(self, change_id: int, **kwargs) -> MemberDeviceChange:
        """
        Retrieves a device change of a member given the change ID
        
        :calls: "GET /members/{id}/device/changes/{changeId}" \
		<https://fabman.io/api/v1/documentation#/members/getMembersIdDeviceChangesChangeId>
  
        :param change_id: ID of the change to retrieve
        :type change_id: int
        :returns: Device change information
        :rtype: dict
        """
        response = self._requester.request(
            "GET",
            f"/members/{self.id}/device/changes/{change_id}",
            _kwargs=kwargs,
        )

        data = response.json()
        data.update({"member_id": self.id})

        return MemberDeviceChange(self._requester, data)

    def get_export(self, **kwargs) -> requests.Response:
        """
        Retrieves the export of a member. This is a placeholder for future functionality.
        
        :calls: "GET /members/{id}/export" \
		<https://fabman.io/api/v1/documentation#/members/getMembersIdExport>
  
        :raises: NotImplementedError
        """
        raise NotImplementedError("get_export not implemented yet")

    def get_invitation(self, **kwargs) -> MemberInvitation:
        """
        Retrieves the invitation status of a member
        
        :calls: "GET /members/{id}/invitations" \
		<https://fabman.io/api/v1/documentation#/members/getMembersIdInvitation>
  
        :returns: Invitation status of a member
        :rtype: dict
        """
        response = self._requester.request(
            "GET",
            f"/members/{self.id}/invitation",
            _kwargs=kwargs,
        )

        data = response.json()
        data.update({"member_id": self.id})

        return MemberInvitation(self._requester, data)

    def get_key(self, **kwargs) -> MemberKey:
        """
        Retrieves the keycard number of a member
        
        :calls: "GET /members/{id}/key" \
		<https://fabman.io/api/v1/documentation#/members/getMembersIdKey>
  
        :returns: :code:`fabman.member.MemberKey` object with key details
        :rtype: fabman.member.MemberKey
        """
        if "key" in self._embedded:
            data = self._embedded["key"]
        else:
            response = self._requester.request(
                "GET",
                f"/members/{self.id}/key",
                _kwargs=kwargs,
            )
            data = response.json()
        data.update({"member_id": self.id})

        return MemberKey(self._requester, data)

    def get_packages(self, **kwargs) -> Union[EmbeddedList, PaginatedList]:
        """
        Retrieves the packages of a member

        :calls: "GET /members/{id}/packages" \
		<https://fabman.io/api/v1/documentation#/members/getMembersIdPackages>
  
        :returns: List of packages of a member
        :rtype: fabman.paginated_list.PaginatedList
        """
        if "memberPackages" in self._embedded:
            return EmbeddedList(
                MemberPackage,
                self._embedded["memberPackages"],
                self._requester,
                "GET",
                f"/members/{self.id}/packages",
                extra_attribs={"member_id": self.id},
            )

        return PaginatedList(
            MemberPackage,
            self._requester,
            "GET",
            f"/members/{self.id}/packages",
            extra_attribs={"member_id": self.id},
            **kwargs,
        )

    def get_package(self, member_package_id: int, **kwargs) -> MemberPackage:
        """
        Retrieves a package of a member
        
        :calls: "GET /members/{id}/packages/{memberPackageId}" \
		<https://fabman.io/api/v1/documentation#/members/getMembersIdPackagesMemberpackageid>
  
        :param member_package_id: ID of the package to retrieve
        :type member_package_id: int
        :returns: :code:`fabman.member.MemberPackage` object with package details
        :rtype: fabman.member.MemberPackage
        """
        data = dict()
        if "memberPackages" in self._embedded:
            for package in self._embedded["memberPackages"]:
                if package["id"] == member_package_id:
                    data = package
        if len(data) == 0:
            response = self._requester.request(
                "GET",
                f"/members/{self.id}/packages/{member_package_id}",
                _kwargs=kwargs,
            )

            data = response.json()
        data.update({"member_id": self.id})

        return MemberPackage(self._requester, data)

    def get_payment_account(self, **kwargs) -> MemberPaymentAccount:
        """
        Retrieves the payment account of a member
        
        :calls: "GET /members/{id}/payment-account" \
		<https://fabman.io/api/v1/documentation#/members/getMembersIdPaymentaccount>
  
        :returns: Payment account information
        :rtype: dict
        """
        response = self._requester.request(
            "GET",
            f"/members/{self.id}/payment-account",
            _kwargs=kwargs,
        )

        data = {"payments": response.json()}
        data.update({"member_id": self.id})

        return MemberPaymentAccount(self._requester, data)

    def get_payment_method(self, **kwargs) -> MemberPaymentMethod:
        """
        Retrieves the payment method of a member
        :calls: "GET /members/{id}/payment-method" \
		<https://fabman.io/api/v1/documentation#/members/getMembersIdPaymentmethod>
  
        :returns: Payment method information
        :rtype: dict
        """
        response = self._requester.request(
            "GET",
            f"/members/{self.id}/payment-method",
            _kwargs=kwargs,
        )

        data = response.json()
        data.update({"member_id": self.id})

        return MemberPaymentMethod(self._requester, data)

    def get_payment_method_mandate_preview(
        self, **kwargs
    ) -> MemberPaymentMethodMandatePreview:
        """
        Retrieves the payment method mandate preview of a member
        
        :calls: "GET /members/{id}/payment-method-mandate-preview" \
		<https://fabman.io/api/v1/documentation#/members/getMembersIdPaymentmethodmandatepreview>
  
        :returns: Payment method mandate preview information
        :rtype: dict
        """
        response = self._requester.request(
            "GET",
            f"/members/{self.id}/payment-method-mandate-preview",
            _kwargs=kwargs,
        )

        data = response.json()
        data.update({"member_id": self.id})

        return MemberPaymentMethodMandatePreview(self._requester, data)

    def get_privileges(self, **kwargs) -> MemberPrivileges:
        """
        Retrieves the privileges of a member
        :calls: "GET /members/{id}/privileges" \
		<https://fabman.io/api/v1/documentation#/members/getMembersIdPrivileges>
  
        :returns: privileges of a member
        :rtype: dict
        """
        if "privileges" in self._embedded:
            data = self._embedded["privileges"]
        else:
            response = self._requester.request(
                "GET",
                f"/members/{self.id}/privileges",
                _kwargs=kwargs,
            )

            data = response.json()
        data.update({"member_id": self.id})

        return MemberPrivileges(self._requester, data)

    def get_trained_resources(self, **kwargs) -> MemberTrainedResources:
        """
        Retrieves the trained resources of a member
        :calls: "GET /members/{id}/trained-resources" \
		<https://fabman.io/api/v1/documentation#/members/getMembersIdTrainedResources>
  
        :return: Trained resources of a member
        :rtype: list
        """
        response = self._requester.request(
            "GET",
            f"/members/{self.id}/trained-resources",
            _kwargs=kwargs,
        )

        data = {"resources": response.json()}
        data.update({"member_id": self.id})

        return MemberTrainedResources(self._requester, data)

    def get_trainings(
        self, ignore_embeds: Optional[bool] = False, **kwargs
    ) -> Union[PaginatedList, EmbeddedList]:
        """
        
        Retrieves the trainings of a member
        :calls: "GET /members/{id}/trainings" \
		<https://fabman.io/api/v1/documentation#/members/getMembersIdTrainings>

        :return: List of trainings of a member
        :rtype: Union[PaginatedList, EmbeddedList]
        """
        if "trainings" in self._embedded and not ignore_embeds:
            return EmbeddedList(
                MemberTraining,
                self._embedded["trainings"],
                self._requester,
                "GET",
                f"/members/{self.id}/trainings",
                extra_attribs={"member_id": self.id},
            )

        return PaginatedList(
            MemberTraining,
            self._requester,
            "GET",
            f"/members/{self.id}/trainings",
            extra_attribs={"member_id": self.id},
        )

    def get_training(self, training_id: int, **kwargs) -> MemberTraining:
        """
        Retrieves a training of a member
        :calls: "GET /members/{id}/trainings/{trainingId}" \
		<https://fabman.io/api/v1/documentation#/members/getMembersIdTrainingstrainingid>
        
        :param training_id: ID of the training to retrieve
        :type training_id: int
        :return: Training information
        :rtype: dict
        """
        data = {}
        if "trainings" in self._embedded:
            for training in self._embedded["trainings"]:
                if training["id"] == training_id:
                    data = training
        if len(data) == 0:
            response = self._requester.request(
                "GET",
                f"/members/{self.id}/trainings/{training_id}",
                _kwargs=kwargs,
            )

            data = response.json()
        data.update({"member_id": self.id})

        return MemberTraining(self._requester, data)

    def refresh(self) -> None:
        """
        Updates the objects with more recent data from the API. Needs to be called
        when update() fails for lockVersioning.

        :calls: "GET /members/{id}" \
		<https://fabman.io/api/v1/documentation#/members/getMembersId>
  
        :returns: None -- updates attributes of the current object
        :rtype: None
        """

        response = self._requester.request("GET", f"/members/{self.id}")

        data = response.json()

        for attr, val in data.items():
            setattr(self, attr, val)

    def update(self, **kwargs) -> None:
        """
        Updates the member object and sets the modified attributes based on what
        is returned by the server.  Member object is updated in place.

        :calls: "PUT /members/{id}" \
		<https://fabman.io/api/v1/documentation#/members/putMembersId>
  
        :returns: None -- updates attributes of the current object
        :rtype: None
        """

        kwargs.update({"lockVersion": self.lockVersion})

        response = self._requester.request(
            "PUT",
            f"/members/{self.id}",
            _kwargs=kwargs,
        )

        data = response.json()

        for attr, val in data.items():
            setattr(self, attr, val)



"""General Utility Functions to be used throughout the package"""

from typing import Union

from requests.structures import CaseInsensitiveDict


def clean_headers(headers: Union[dict, CaseInsensitiveDict]):
    """Cleans the headers to hide sensitive information in logs.
    Originally defined in canvasapi/util.py:
    https://github.com/ucfopen/canvasapi/blob/develop/canvasapi/util.py#L228

    Args:
        headers (dict): The headers to sanitize

    Returns:
        _type_: _description_
    """
    cleaned_headers = headers.copy()

    authorization_header = headers.get("Authorization")
    if authorization_header:
        sanitized = "****" + authorization_header[-4:]
        cleaned_headers["Authorization"] = sanitized

    return cleaned_headers



"""Tests for the Member class."""
# pylint: disable=missing-function-docstring, missing-class-docstring, invalid-name, unused-argument, protected-access

import unittest
import warnings

import requests_mock

from fabman import Fabman
from fabman.embedded_list import EmbeddedList
from fabman.exceptions import ResourceDoesNotExist
from fabman.member import (
    Member,
    MemberBalanceItems,
    MemberChange,
    MemberCredit,
    MemberCreditUse,
    MemberDevice,
    MemberDeviceChange,
    MemberInvitation,
    MemberKey,
    MemberPackage,
    MemberPaymentAccount,
    MemberPaymentMethod,
    MemberPaymentMethodMandatePreview,
    MemberPrivileges,
    MemberTrainedResources,
    MemberTraining,
)
from fabman.package import Package
from fabman.paginated_list import PaginatedList
from tests import settings
from tests.util import register_uris, validate_update


@requests_mock.Mocker()
class TestMembers(unittest.TestCase):
    def setUp(self):
        self.fabman = Fabman(settings.API_KEY)

        with requests_mock.Mocker() as m:
            register_uris({"fabman": ["get_member_by_id"]}, m)

            self.member: Member = self.fabman.get_member(1)

    def test_sanity(self, m):
        self.assertEqual(1, 1)

    def test_to_str(self, m):
        string = str(self.member)
        self.assertIsInstance(string, str)
        self.assertTrue("1: Julian Bashear" == string)

    def test_create_credit(self, m):
        register_uris({"member": ["create_credit"]}, m)

        credit = self.member.create_credit(amount=12.34)
        self.assertIsInstance(credit, MemberCredit)
        self.assertTrue(hasattr(credit, "amount"))
        self.assertTrue(credit.amount == 12.34)

    def test_create_key(self, m):
        register_uris({"member": ["create_key"]}, m)

        key = self.member.create_key(type="em4102", token="12345678")
        self.assertIsInstance(key, MemberKey)
        self.assertTrue(hasattr(key, "type"))
        self.assertTrue(key.type == "em4102")

    def test_create_training(self, m):
        register_uris({"member": ["create_training"]}, m)

        training = self.member.create_training(account="1", training_course="1")
        self.assertIsInstance(training, MemberTraining)
        self.assertTrue(hasattr(training, "id"))

    def test_delete(self, m):
        register_uris({"member": ["delete"]}, m)

        resp = self.member.delete()
        self.assertTrue(m.called)
        self.assertTrue(resp.status_code == 204)

    def test_get_balance_items(self, m):
        register_uris({"member": ["get_balance_items"]}, m)

        balance_items = self.member.get_balance_items()

        self.assertIsInstance(balance_items, MemberBalanceItems)

    def test_get_get_changes(self, m):
        register_uris({"member": ["get_changes"]}, m)

        changes = self.member.get_changes()
        self.assertIsInstance(changes, list)

    def test_get_credits(self, m):
        register_uris({"member": ["get_credits"]}, m)

        _credits = self.member.get_credits()
        self.assertIsInstance(_credits, PaginatedList)

    def test_get_credit_by_credit_id(self, m):
        register_uris({"member": ["get_credit_by_id"]}, m)

        credit = self.member.get_credit(1)
        self.assertIsInstance(credit, MemberCredit)

    def test_get_device(self, m):
        register_uris({"member": ["get_device"]}, m)

        device = self.member.get_device()
        self.assertIsInstance(device, MemberDevice)
        self.assertEqual(device.name, "Starfleet Tricorder")

        self.assertEqual(str(device), "Device: Starfleet Tricorder")

    def test_get_device_embeded(self, m):
        register_uris({"fabman": ["get_member_by_id_device_embed"]}, m)

        member = self.fabman.get_member(1, embed=["device"])

        self.assertIsInstance(member, Member)
        device = member.get_device()
        self.assertTrue(device.name == "Device 1")

    def test_get_device_changes(self, m):
        register_uris({"member": ["get_device_changes"]}, m)
        changes = self.member.get_device_changes()

        self.assertIsInstance(changes, list)
        self.assertIsInstance(changes[0], MemberDeviceChange)
        self.assertEqual(len(changes), 2)

    def test_get_device_changes_by_id(self, m):
        register_uris({"member": ["get_device_change_by_id"]}, m)

        change = self.member.get_device_change(1)

        self.assertIsInstance(change, MemberDeviceChange)

    def test_get_embeds(self, m):
        register_uris({"member": ["get_embeds"]}, m)

        member = self.fabman.get_member(
            1, embed=["memberPackages", "trainings", "key", "privileges", "device"]
        )
        self.assertIsInstance(member, Member)
        self.assertTrue(hasattr(member, "_embedded"))
        self.assertTrue("memberPackages" in member._embedded)
        self.assertTrue("trainings" in member._embedded)
        self.assertTrue("key" in member._embedded)
        self.assertTrue("privileges" in member._embedded)
        self.assertTrue("device" in member._embedded)

        # memberPackages
        packages = member.get_packages()
        self.assertIsInstance(packages, EmbeddedList)
        self.assertIsInstance(packages[0], MemberPackage)
        self.assertEqual(packages[0].member_id, 1)

        package = member.get_package(1)
        self.assertIsInstance(package, MemberPackage)
        self.assertEqual(package.member_id, 1)

        # trainings
        trainings = member.get_trainings()
        self.assertIsInstance(trainings, EmbeddedList)
        self.assertIsInstance(trainings[0], MemberTraining)
        self.assertTrue(trainings[0].id == 2)
        self.assertTrue(trainings[0].member_id == 1)

        training = member.get_training(2)
        self.assertIsInstance(training, MemberTraining)
        self.assertTrue(training.id == 2)
        self.assertTrue(training.member_id == 1)

        # key
        key = member.get_key()
        self.assertIsInstance(key, MemberKey)
        self.assertTrue(key.lockVersion == 4)
        self.assertTrue(key.member_id == 1)

        # privileges
        privileges = member.get_privileges()
        self.assertIsInstance(privileges, MemberPrivileges)
        self.assertTrue(privileges.privileges == "member")
        self.assertEqual(privileges.member_id, 1)

        # device
        device = member.get_device()
        self.assertIsInstance(device, MemberDevice)
        self.assertTrue(device.name == "Starfleet Tricorder")

    def test_get_invitation(self, m):
        register_uris({"member": ["get_invitation"]}, m)

        invitation = self.member.get_invitation()
        self.assertIsInstance(invitation, MemberInvitation)
        self.assertEqual(invitation.member_id, 1)

    def test_get_key_empty(self, m):
        register_uris({"member": ["get_key_empty"]}, m)
        with warnings.catch_warnings(record=True):
            key = self.member.get_key()  # pylint: disable=unused-variable
            self.assertRaises(
                UserWarning,
                msg=(
                    "204 No Content returned. This likely means there is no information"
                    "at the resource."
                ),
            )

    def test_get_key(self, m):
        register_uris({"member": ["get_key_data"]}, m)
        key = self.member.get_key()

        self.assertIsInstance(key, MemberKey)
        self.assertTrue(key.type == "em4102")

    def test_get_packages(self, m):
        register_uris({"member": ["get_packages"]}, m)
        key = self.member.get_packages()

        self.assertIsInstance(key, PaginatedList)

    def test_get_package_by_id(self, m):
        register_uris({"member": ["get_package_by_id"]}, m)
        key = self.member.get_package(1)

        self.assertIsInstance(key, MemberPackage)

    def test_get_package_by_id_doesnt_hold(self, m):
        register_uris({"member": ["get_package_by_id_doesnt_hold"]}, m)

        self.assertRaises(ResourceDoesNotExist, self.member.get_package, 2)

    def test_get_payment_account(self, m):
        register_uris({"member": ["get_payment_account"]}, m)

        payment_acct = self.member.get_payment_account()

        self.assertIsInstance(payment_acct, MemberPaymentAccount)
        self.assertEqual(payment_acct.member_id, 1)

    def test_get_payment_methods_exists(self, m):
        register_uris({"member": ["get_payment_method_exists"]}, m)

        payment_method = self.member.get_payment_method()
        self.assertIsInstance(payment_method, MemberPaymentMethod)
        self.assertEqual(payment_method.member_id, 1)

    def test_get_payment_method_doesnt_exist(self, m):
        register_uris({"member": ["get_payment_method_doesnt_exist"]}, m)
        with warnings.catch_warnings(record=True):  # pylint: disable=unused-variable
            self.assertRaises(
                UserWarning,
                msg=(
                    "204 No Content returned. This likely means there is no information"
                    "at the resource."
                ),
            )

    def test_get_payment_method_mandate_preview(self, m):
        register_uris({"member": ["get_payment_method_mandate_preview"]}, m)

        mandate_preview = self.member.get_payment_method_mandate_preview()
        self.assertIsInstance(mandate_preview, MemberPaymentMethodMandatePreview)
        self.assertEqual(mandate_preview.member_id, 1)

    def test_get_privileges(self, m):
        register_uris({"member": ["get_privileges"]}, m)

        priv = self.member.get_privileges()
        self.assertIsInstance(priv, MemberPrivileges)
        self.assertEqual(priv.privileges, "admin")
        self.assertEqual(priv.member_id, 1)
        self.assertTrue(str(priv) == "Member #1 is admin")

    def test_get_trained_resources(self, m):
        register_uris({"member": ["get_trained_resources"]}, m)

        train_res = self.member.get_trained_resources()
        self.assertIsInstance(train_res, MemberTrainedResources)
        self.assertEqual(train_res.member_id, 1)
        self.assertTrue(str(train_res) == "Member #1 is trained on 3 resources")
        self.assertListEqual(train_res.resources, [1, 2, 3])

    def test_get_trainings(self, m):
        register_uris({"member": ["get_trainings"]}, m)

        trainings = self.member.get_trainings()
        self.assertIsInstance(trainings, PaginatedList)
        self.assertIsInstance(trainings[0], MemberTraining)

    def test_get_training_by_id(self, m):
        register_uris({"member": ["get_training_by_id"]}, m)

        training = self.member.get_training(1)
        self.assertIsInstance(training, MemberTraining)

    def test_refresh(self, m):
        register_uris({"fabman": ["get_member_by_id"]}, m)

        before = self.member.id
        self.member.refresh()
        self.assertTrue(m.called)
        self.assertTrue(before == self.member.id)

    def test_update(self, m):
        m.register_uri(
            "PUT",
            "https://fabman.io/api/v1/members/1",
            text=validate_update,
            status_code=200,
        )

        self.member.update(firstName="John", lastName="Doe")
        self.assertTrue(m.called)


@requests_mock.Mocker()
class TestMemberChange(unittest.TestCase):
    def setUp(self):
        self.fabman = Fabman(settings.API_KEY)

        with requests_mock.Mocker() as m:
            register_uris({"fabman": ["get_member_by_id"]}, m)
            register_uris({"member": ["get_changes"]}, m)

            self.member: Member = self.fabman.get_member(1)
            self.changes: list = self.member.get_changes()
            self.change = self.changes[0]

    def test_instance(self, m):
        self.assertIsInstance(self.change, MemberChange)

    def test_string(self, m):
        self.assertEqual(str(self.change), "Change #1 for member #1")

    def test_delete(self, m):
        register_uris({"member": ["delete_change"]}, m)

        resp = self.change.delete()
        self.assertTrue(m.called)
        self.assertTrue(resp.status_code == 204)


@requests_mock.Mocker()
class TestMemberCredit(unittest.TestCase):
    def setUp(self):
        self.fabman = Fabman(settings.API_KEY)

        with requests_mock.Mocker() as m:
            register_uris({"fabman": ["get_member_by_id"]}, m)
            register_uris({"member": ["get_credit_by_id"]}, m)

            self.member: Member = self.fabman.get_member(1)
            self.credit: MemberCredit = self.member.get_credit(1)

    def test_member_id(self, m):
        self.assertTrue(hasattr(self.credit, "member_id"))

    def test_to_str(self, m):
        string = str(self.credit)
        self.assertIsInstance(string, str)
        self.assertTrue("1: global - 12.34" == string)

    def test_delete(self, m):
        register_uris({"member": ["delete_credit"]}, m)

        resp = self.credit.delete()
        self.assertTrue(m.called)
        self.assertTrue(resp.status_code == 204)

    def test_get_uses(self, m):
        register_uris({"member": ["get_credit_uses"]}, m)

        uses = self.credit.get_uses()
        self.assertIsInstance(uses, list)
        self.assertIsInstance(uses[0], MemberCreditUse)
        self.assertTrue(uses[0].id == 1)

    def test_update(self, m):
        m.register_uri(
            "PUT",
            "https://fabman.io/api/v1/members/1/credits/1",
            text=validate_update,
            status_code=200,
        )

        self.credit.update(amount=12.34)
        self.assertTrue(m.called)


@requests_mock.Mocker()
class TestMemberDeviceChange(unittest.TestCase):
    def setUp(self):
        self.fabman = Fabman(settings.API_KEY)

        with requests_mock.Mocker() as m:
            register_uris({"fabman": ["get_member_by_id"]}, m)
            register_uris({"member": ["get_device_changes"]}, m)

            self.member: Member = self.fabman.get_member(1)
            self.device_changes: list = self.member.get_device_changes()
            self.device_change = self.device_changes[0]

    def test_instance(self, m):
        self.assertIsInstance(self.device_change, MemberDeviceChange)

    def test_str(self, m):
        self.assertEqual(str(self.device_change), "DeviceChange #1 for member #1")

    def test_delete(self, m):
        register_uris({"member": ["delete_device_change"]}, m)

        resp = self.device_change.delete()
        self.assertTrue(m.called)
        self.assertTrue(resp.status_code == 204)


@requests_mock.Mocker()
class TestMemberPaymentMethod(unittest.TestCase):
    def setUp(self):
        self.fabman = Fabman(settings.API_KEY)

        with requests_mock.Mocker() as m:
            register_uris({"fabman": ["get_member_by_id"]}, m)
            register_uris({"member": ["get_payment_method_exists"]}, m)

            self.member: Member = self.fabman.get_member(1)
            self.payment_method: MemberPaymentMethod = self.member.get_payment_method()

    def test_instance(self, m):
        self.assertIsInstance(self.payment_method, MemberPaymentMethod)

    def test_str(self, m):
        self.assertEqual(
            str(self.payment_method), "Member #1 has a stripe payment method"
        )

    def test_delete(self, m):
        register_uris({"member": ["delete_payment_method"]}, m)

        resp = self.payment_method.delete()
        self.assertTrue(m.called)
        self.assertTrue(resp.status_code == 204)


@requests_mock.Mocker()
class TestMemberKey(unittest.TestCase):
    def setUp(self):
        self.fabman: Fabman = Fabman(settings.API_KEY)

        with requests_mock.Mocker() as m:
            register_uris(
                {"member": ["get_key_data"], "fabman": ["get_member_by_id"]}, m
            )
            self.key: MemberKey = self.fabman.get_member(1).get_key()

    def test_to_str(self, m):
        string = str(self.key)
        self.assertIsInstance(string, str)
        self.assertTrue("1 - em4102" == string)

    def test_delete(self, m):
        register_uris({"member": ["delete_key"]}, m)

        resp = self.key.delete()
        self.assertTrue(m.called)
        self.assertTrue(resp.status_code == 204)

    def test_update(self, m):
        m.register_uri(
            "PUT",
            f"{settings.BASE_URL_WITH_VERSION}/members/1/key",
            text=validate_update,
            status_code=200,
        )

        self.key.update(type="blowfish")  # the reallest key
        self.assertTrue(m.called)


@requests_mock.Mocker()
class TestMemberPackage(unittest.TestCase):
    def setUp(self):
        self.fabman = Fabman(settings.API_KEY)

        with requests_mock.Mocker() as m:
            register_uris(
                {"fabman": ["get_member_by_id"], "member": ["get_package_by_id"]}, m
            )

            self.member = self.fabman.get_member(1)
            self.package = self.member.get_package(1)

    def test_str(self, m):
        string = str(self.package)
        self.assertIsInstance(string, str)
        self.assertTrue("1 - 1" == string)

    def test_delete(self, m):
        register_uris({"member": ["delete_package"]}, m)

        resp = self.package.delete()
        self.assertTrue(m.called)
        self.assertTrue(resp.status_code == 204)

    def test_get_package(self, m):
        register_uris(
            {"member": ["get_package_by_id"], "fabman": ["get_package_by_id"]}, m
        )

        package = self.package.get_package()
        self.assertIsInstance(package, Package)

    def test_get_package_embedded(self, m):
        register_uris({"member": ["get_packages_embedded"]}, m)

        memberPackage = self.member.get_packages(embed=["package"])
        self.assertIsInstance(memberPackage, PaginatedList)
        self.assertIsInstance(memberPackage[0], MemberPackage)

        package = memberPackage[0].get_package()
        self.assertIsInstance(package, Package)
        self.assertTrue(hasattr(package, "name"))
        self.assertTrue(package.name == "Test Package")

    def test_update(self, m):
        m.register_uri(
            "PUT",
            "https://fabman.io/api/v1/members/1/packages/1",
            text=validate_update,
            status_code=200,
        )

        self.package.update()
        self.assertTrue(m.called)


@requests_mock.Mocker()
class TestMemberTraining(unittest.TestCase):
    def setUp(self) -> None:
        self.fabman = Fabman(settings.API_KEY)

        with requests_mock.Mocker() as m:
            register_uris(
                {"fabman": ["get_member_by_id"], "member": ["get_training_by_id"]}, m
            )

            self.member = self.fabman.get_member(1)
            self.training = self.member.get_training(1)

    def test_instance(self, m):
        self.assertIsInstance(self.training, MemberTraining)

    def test_str(self, m):
        self.assertTrue(str(self.training) == "MemberTraining #1 for member #1")

    def test_delete(self, m):
        register_uris({"member": ["delete_training"]}, m)

        resp = self.training.delete()
        self.assertTrue(m.called)
        self.assertTrue(resp.status_code == 204)



"""Tests for the Payment class."""
# pylint: disable=missing-docstring, invalid-name, unused-argument

import unittest

import requests
import requests_mock

from fabman import Fabman
from fabman.payment import Payment, PaymentRequest
from tests import settings
from tests.util import register_uris, validate_update


@requests_mock.Mocker()
class TestPayment(unittest.TestCase):
    """
    Test cases for the Payment class. Note that the developers of this library
    are cheap and have not validated against real data. Sample data is taken from
    the Fabman API POST command model.
    """

    def setUp(self) -> None:
        self.fabman = Fabman(settings.API_KEY)

        with requests_mock.Mocker() as m:
            register_uris({"fabman": ["get_payment_by_id"]}, m)

            self.payment: Payment = self.fabman.get_payment(1)

    def test_instance(self, m):
        """Tests to make sure the constructor worked as expected"""
        self.assertIsInstance(self.payment, Payment)

    def test_str(self, m):
        string = str(self.payment)
        self.assertTrue(string == "Payment #1: Test Payment")

    def test_delete(self, m):
        register_uris({"payment": ["delete"]}, m)

        resp = self.payment.delete()
        self.assertTrue(m.called)
        self.assertIsInstance(resp, requests.Response)
        self.assertTrue(resp.status_code == 204)

    def test_request_payment(self, m):
        register_uris({"payment": ["request_payment"]}, m)

        resp = self.payment.request_payment()
        self.assertTrue(m.called)
        self.assertIsInstance(resp, PaymentRequest)
        self.assertTrue(resp.payment_id == 1)

        self.assertEqual(str(resp), "PaymentRequest for Payment #1")

    def test_update(self, m):
        """Checks to ensure lockVersion is added and used"""
        m.register_uri(
            "PUT",
            f"{settings.BASE_URL_WITH_VERSION}/payments/1",
            text=validate_update,
            status_code=200,
        )

        self.payment.update(notes="Updated Notes")
        self.assertTrue(m.called)



"""Defines and handles the TrainingCourse object"""
import requests

from fabman.fabman_object import FabmanObject


class TrainingCourse(FabmanObject):
    """TrainingCourse Object handles all API calls that operate on a single TrainingCourse."""

    def __str__(self):
        return f"TrainingCourse #{self.id}: {self.title}"

    def delete(self, **kwargs) -> requests.Response:
        """
        Delete the training course. *WARNING: THIS CANNOT BE UNDONE.*

        :calls: "DELETE /training-courses/{id}" \
		<https://fabman.io/api/v1/documentation#/training-courses/deleteTrainingcoursesId>
  
        :return: response information of call
        :rtype: requests.Response
        """
        uri = f"/training-courses/{self.id}"

        response = self._requester.request("DELETE", uri, _kwargs=kwargs)

        return response

    def update(self, **kwargs) -> None:
        """
        Update the training course with the given data. Will create the data
        if the training course does not exist.

        :calls: "PUT /training-courses/{id}" \
		<https://fabman.io/api/v1/documentation#/training-courses/putTrainingcoursesId>
  
        :return: None -- attributes are updated in place
        :rtype: None
        """
        uri = f"/training-courses/{self.id}"

        kwargs.update({"lockVersion": self.lockVersion})
        response = self._requester.request("PUT", uri, _kwargs=kwargs)

        data = response.json()

        for attr, val in data.items():
            setattr(self, attr, val)



"""Defines the Webhook class"""

import requests

from fabman.fabman_object import FabmanObject


class Webhook(FabmanObject):
    """
    Class for interacting with the webhooks endpoint on the Fabman API
    """

    def __str__(self):
        return f"Webhook #{self.id}: {self.label}"

    def delete(self, **kwargs):
        """
        Deletes a webhook. *WARNING: This is irreversible.*

        :calls: "DELETE /webhooks/{webhookId}" \
		<https://fabman.io/api/v1/documentation#/webhooks/deleteWebhooksId>
  
        :return: response information of call
        :rtype: requests.Response
        """

        uri = f"/webhooks/{self.id}"

        response = self._requester.request("DELETE", uri, _kwargs=kwargs)

        return response

    def get_events(self, **kwargs) -> requests.Response:
        """
        Returns the events for a webhook.

        :calls: "GET /webhooks/{webhookId}/events" \
		<https://fabman.io/api/v1/documentation#/webhooks/getWebhooksIdEvents>
  
        :return: response information of call
        :rtype: requests.Response
        """

        if "events" in self._embedded:
            return self._embedded["events"]

        uri = f"/webhooks/{self.id}/events"

        response = self._requester.request("GET", uri, _kwargs=kwargs)

        return response

    def send_test_event(self, **kwargs) -> requests.Response:
        """
        Sends a test event to the webhook.

        :calls: "POST /webhooks/{webhookId}/events" \
		<https://fabman.io/api/v1/documentation#/webhooks/postWebhooksIdTest>
  
        :return: response information of call
        :rtype: requests.Response
        """

        uri = f"/webhooks/{self.id}/test"

        response = self._requester.request("POST", uri, _kwargs=kwargs)

        return response

    def update(self, **kwargs) -> None:
        """
        Updates the webhook. Attributes are updated in place with new information

        :calls: "PUT /webhooks/{webhookId}" \
		<https://fabman.io/api/v1/documentation#/webhooks/putWebhooksId>
  
        :return: None -- attributes are updated in place
        :rtype: None
        """

        uri = f"/webhooks/{self.id}"

        kwargs.update({"lockVersion": self.lockVersion})
        response = self._requester.request("PUT", uri, _kwargs=kwargs)

        data = response.json()

        for attr, val in data.items():
            setattr(self, attr, val)



"""Tests for Booking class."""
# pylint: disable=missing-docstring, invalid-name, unused-argument

import unittest

import requests_mock

from fabman import Fabman
from fabman.booking import Booking
from tests import settings
from tests.util import register_uris, validate_update


@requests_mock.Mocker()
class TestBooking(unittest.TestCase):
    def setUp(self):
        self.fabman = Fabman(settings.API_KEY)

        with requests_mock.Mocker() as m:
            register_uris({"fabman": ["get_booking_by_id"]}, m)

            self.booking: Booking = self.fabman.get_booking(1)

    def test_instance(self, m):
        self.assertIsInstance(self.booking, Booking)

    def test_str(self, m):
        string = str(self.booking)
        self.assertTrue(string == "Booking #1: 2023-01-01T14:00 - 2023-01-01T15:00")

    def test_delete(self, m):
        register_uris({"booking": ["delete"]}, m)

        resp = self.booking.delete()
        self.assertTrue(m.called)
        self.assertTrue(resp.status_code == 204)

    def test_update(self, m):
        m.register_uri(
            "PUT",
            f"{settings.BASE_URL_WITH_VERSION}/bookings/1",
            text=validate_update,
            status_code=200,
        )

        self.booking.update()
        self.assertTrue(m.called)
